# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LensKit pipeline builder.
"""

# pyright: strict
from __future__ import annotations

import typing
import warnings
from copy import deepcopy
from graphlib import CycleError, TopologicalSorter
from os import PathLike
from pathlib import Path
from types import UnionType
from uuid import NAMESPACE_URL, uuid5

from typing_extensions import Any, Literal, TypeVar, cast, overload

from lenskit.config import load_config_data
from lenskit.diagnostics import PipelineError, PipelineWarning
from lenskit.logging import get_logger

from . import config
from ._hooks import ComponentInputHook, HookEntry, RunHooks
from ._impl import Pipeline
from .cache import PipelineCache
from .components import (
    Component,
    ComponentConstructor,
    PipelineFunction,
    fallback_on_none,
)
from .config import PipelineConfig, PipelineHook
from .nodes import (
    ND,
    ComponentConstructorNode,
    ComponentInstanceNode,
    ComponentNode,
    InputNode,
    LiteralNode,
    Node,
)
from .types import TypecheckWarning, resolve_type_string

_log = get_logger(__name__)

CFG = TypeVar("CFG")
# common type var for quick use
T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")

NAMESPACE_LITERAL_DATA = uuid5(NAMESPACE_URL, "https://ns.lenskit.org/literal-data/")


class PipelineBuilder:
    """
    Builder for LensKit recommendation pipelines.  :ref:`Pipelines <pipeline>`
    are the core abstraction for using LensKit models and other components to
    produce recommendations in a useful way.  They allow you to wire together
    components in (mostly) abitrary graphs, train them on data, and serialize
    the resulting pipelines to disk for use elsewhere.

    The builder configures and builds pipelines that can then be run. If you
    have a scoring model and just want to generate recommenations with a default
    setup and minimal configuration, see :func:`~lenskit.pipeline.topn_pipeline`
    or :class:`~lenskit.pipeline.RecPipelineBuilder`.

    Args:
        name:
            A name for the pipeline.
        version:
            A numeric version for the pipeline.

    Stability:
        Caller
    """

    name: str | None = None
    """
    The pipeline name.
    """
    version: str | None = None
    """
    The pipeline version string.
    """

    _nodes: dict[str, Node[Any]]
    _edges: dict[str, dict[str, str]]
    _aliases: dict[str, Node[Any]]
    _default_connections: dict[str, str]
    _default: str | None = None
    _run_hooks: dict[str, list[HookEntry[Any]]]

    def __init__(self, name: str | None = None, version: str | None = None):
        self.name = name
        self.version = version
        self._nodes = {}
        self._edges = {}
        self._aliases = {}
        self._default_connections = {}
        self._run_hooks = {}

    @classmethod
    def from_pipeline(cls, pipeline: Pipeline) -> PipelineBuilder:
        """
        Create a builder initialized with a pipeline's internal state.  See
        :meth:`Pipeline.modify` for details — that is the main entry point, and
        this method exists to be the implementation of that method.
        """
        builder = cls()
        builder._nodes = {n.name: n for n in pipeline.nodes()}
        builder._aliases = {a: builder.node(t) for (a, t) in pipeline.config.aliases.items()}
        for name, spec in pipeline.config.components.items():
            builder._edges[name] = spec.inputs
        builder._default = pipeline.config.default
        builder._run_hooks = {
            name: [h for h in hooks if h.priority != 0]  # type: ignore
            for name, hooks in pipeline._run_hooks.items()  # type: ignore
        }

        return builder

    def meta(self, *, include_hash: bool = True) -> config.PipelineMeta:
        """
        Get the metadata (name, version, hash, etc.) for this pipeline without
        returning the whole config.

        Args:
            include_hash:
                Whether to include a configuration hash in the metadata.
        """
        meta = config.PipelineMeta(name=self.name, version=self.version)
        if include_hash:
            meta.hash = self.config_hash()
        return meta

    def nodes(self) -> list[Node[object]]:
        """
        Get the nodes in the pipeline graph.
        """
        return list(self._nodes.values())

    @overload
    def node(self, node: str, *, missing: Literal["error"] = "error") -> Node[object]: ...
    @overload
    def node(self, node: str, *, missing: Literal["none"] | None) -> Node[object] | None: ...
    @overload
    def node(self, node: Node[T]) -> Node[T]: ...
    def node(
        self, node: str | Node[Any], *, missing: Literal["error", "none"] | None = "error"
    ) -> Node[object] | None:
        """
        Get the pipeline node with the specified name.  If passed a node, it
        returns the node or fails if the node is not a member of the pipeline.

        Args:
            node:
                The name of the pipeline node to look up, or a node to check for
                membership.

        Returns:
            The pipeline node, if it exists.

        Raises:
            KeyError:
                The specified node does not exist.
        """
        if isinstance(node, Node):
            self._check_member_node(node)
            return node
        elif node in self._aliases:
            return self._aliases[node]
        elif node in self._nodes:
            return self._nodes[node]
        elif missing == "none" or missing is None:
            return None
        else:
            raise KeyError(f"node {node}")

    def create_input(self, name: str, *types: type[T] | UnionType | None) -> Node[T]:
        """
        Create an input node for the pipeline.  Pipelines expect their inputs to
        be provided when they are run.

        Args:
            name:
                The name of the input.  The name must be unique in the pipeline
                (among both components and inputs).
            types:
                The allowable types of the input; input data can be of any
                specified type.  If ``None`` is among the allowed types, the
                input can be omitted.

        Returns:
            A pipeline node representing this input.

        Raises:
            ValueError:
                a node with the specified ``name`` already exists.
        """
        self._check_available_name(name)

        rts: set[type[T | None]] = set()
        for t in types:
            if t is None:
                rts.add(type(None))
            elif isinstance(t, UnionType):
                rts |= set(typing.get_args(t))
            else:
                rts.add(t)

        node = InputNode[Any](name, types=rts)
        self._nodes[name] = node
        return node

    def literal(self, value: T, *, name: str | None = None) -> LiteralNode[T]:
        """
        Create a literal node (a node with a fixed value).

        .. note::
            Literal nodes cannot be serialized witih :meth:`get_config` or
            :meth:`save_config`.
        """
        if name is None:
            lit = config.PipelineLiteral.represent(value)
            name = str(uuid5(NAMESPACE_LITERAL_DATA, lit.model_dump_json()))
        node = LiteralNode(name, value, types=set([type(value)]))
        self._nodes[name] = node
        return node

    def default_connection(self, name: str, node: Node[Any] | object) -> None:
        """
        Set the default wiring for a component input.  Components that declare
        an input parameter with the specified ``name`` but no configured input
        will be wired to this node.

        This is intended to be used for things like wiring up `user` parameters
        to semi-automatically receive the target user's identity and history.

        .. important::

            Defaults are a feature of the builder only, and are resolved in
            :meth:`build`.  They are not included in serialized configuration or
            resulting pipeline.

        Args:
            name:
                The name of the parameter to set a default for.
            node:
                The node or literal value to wire to this parameter.
        """
        if not isinstance(node, Node):
            node = self.literal(node)
        self._default_connections[name] = node.name

    def default_component(self, node: str | Node[Any]) -> None:
        """
        Set the default node for the pipeline.  If :meth:`Pipeline.run` is
        called without a node, then it will run this node (and all of its
        dependencies).
        """
        self._default = node.name if isinstance(node, Node) else node

    def remove_alias(self, alias: str, *, exist_ok: bool = False):
        """
        Remove an alias from the builder.
        """
        try:
            del self._aliases[alias]
        except KeyError as e:
            if exist_ok:
                pass
            else:
                raise e

    def alias(self, alias: str, node: Node[Any] | str, *, replace: bool = False) -> None:
        """
        Create an alias for a node.  After aliasing, the node can be retrieved
        from :meth:`node` using either its original name or its alias.

        Args:
            alias:
                The alias to add to the node.
            node:
                The node (or node name) to alias.
            replace:
                If ``True``, replace the alias if one already exists.

        Raises:
            ValueError:
                if the alias is already used as an alias or node name.
        """
        node = self.node(node)
        if replace:
            if alias in self._nodes:
                del self._nodes[alias]
        else:
            self._check_available_name(alias)
        self._aliases[alias] = node

    @overload
    def add_component(
        self,
        name: str,
        cls: ComponentConstructor[CFG, ND],
        config: CFG = None,
        /,
        **inputs: Node[Any],
    ) -> Node[ND]: ...
    @overload
    def add_component(
        self,
        name: str,
        instance: Component[ND] | PipelineFunction[ND],
        /,
        **inputs: Node[Any] | object,
    ) -> Node[ND]: ...
    def add_component(
        self,
        name: str,
        comp: ComponentConstructor[CFG, ND] | Component[ND] | PipelineFunction[ND],
        config: CFG | None = None,
        /,
        **inputs: Node[Any] | object,
    ) -> Node[ND]:
        """
        Add a component and connect it into the graph.

        Args:
            name:
                The name of the component in the pipeline.  The name must be
                unique in the pipeline (among both components and inputs).
            cls:
                A component class.
            config:
                The configuration object for the component class.
            instance:
                A raw function or pre-instantiated component.
            inputs:
                The component's input wiring.  See :ref:`pipeline-connections`
                for details.

        Returns:
            The node representing this component in the pipeline.
        """
        self._check_available_name(name)

        node = ComponentNode[ND].create(name, comp, config)
        if node.types is None:
            warnings.warn(f"cannot determine return type of component {comp}", TypecheckWarning)
        self._nodes[name] = node

        self.connect(node, **inputs)

        return node

    @overload
    def replace_component(
        self,
        name: str | Node[ND],
        cls: ComponentConstructor[CFG, ND],
        config: CFG = None,
        /,
        **inputs: Node[Any],
    ) -> Node[ND]: ...
    @overload
    def replace_component(
        self,
        name: str | Node[ND],
        instance: Component[ND] | PipelineFunction[ND],
        /,
        **inputs: Node[Any] | object,
    ) -> Node[ND]: ...
    def replace_component(
        self,
        name: str | Node[ND],
        comp: ComponentConstructor[CFG, ND] | Component[ND] | PipelineFunction[ND],
        config: CFG | None = None,
        /,
        **inputs: Node[Any] | object,
    ) -> Node[ND]:
        """
        Replace a component in the graph.  The new component must have a type
        that is compatible with the old component.  Both input and output connections
        are retained, except for those overridden with with keyword arguments.

        Args:
            name:
                The name or node to replace.
            comp:
                The component or constructor to use instead of the current node's
                component.
            config:
                A configuration for the component (if passed as a class or constructor).
            inputs:
                New input wiring(s) for the new component.
        """
        if isinstance(name, Node):
            name = name.name

        node = ComponentNode[ND].create(name, comp, config)
        self._nodes[name] = node

        self.connect(node, **inputs)

        return node

    def connect(self, obj: str | Node[Any], **inputs: Node[Any] | str | object):
        """
        Provide additional input connections for a component that has already
        been added.  See :ref:`pipeline-connections` for details.

        Args:
            obj:
                The name or node of the component to wire.
            inputs:
                The component's input wiring.  For each keyword argument in the
                component's function signature, that argument can be provided
                here with an input that the pipeline will provide to that
                argument of the component when the pipeline is run.
        """
        if isinstance(obj, Node):
            node = obj
        else:
            node = self.node(obj)
        if not isinstance(node, ComponentNode):
            raise TypeError(f"only component nodes can be wired, not {node}")

        edges = self._edges.get(node.name, None)
        if edges is None:
            self._edges[node.name] = edges = {}

        for k, n in inputs.items():
            if isinstance(n, Node):
                n = cast(Node[Any], n)
                self._check_member_node(n)
                edges[k] = n.name
            else:
                lit = self.literal(n)
                edges[k] = lit.name

    def clear_inputs(self, node: str | Node[Any]):
        """
        Remove input wirings for a node.

        Args:
            node:
                The node whose input wiring should be removed.
        """
        if isinstance(node, Node):
            node = node.name

        self._edges[node] = {}

    def add_run_hook(
        self, name: Literal["component-input"], hook: ComponentInputHook, *, priority: int = 1
    ) -> None:
        """
        Add a hook to be called when the pipeline is run (see :ref:`pipeline-hooks`).

        Args:
            name:
                The name of the hook to add a handler for.
            hook:
                The hook function to run.
            priority:
                The hook priority. Hooks are run in ascending priority, and
                hooks with the same priority are run in the order they are
                added.  LensKit's built-in hooks run at priority 0.
        """
        if priority == 0:
            raise ValueError("priority 0 is reserved for LensKit internal hooks")
        hooks = self._run_hooks.setdefault(name, [])
        hooks.append(HookEntry(hook, priority))

    def validate(self):
        """
        Check the built pipeline for errors.
        """

        # Check for cycles
        graph = {n: set(w.values()) for (n, w) in self._edges.items()}
        ts = TopologicalSorter(graph)
        try:
            ts.prepare()
        except CycleError as e:
            raise PipelineError("pipeline has cycles") from e

    def clone(self) -> PipelineBuilder:
        """
        Clone the pipeline builder.  The resulting builder starts as a copy of
        this builder, and any subsequent modifications only the copy to which
        they are applied.
        """

        clone = PipelineBuilder()

        for node in self.nodes():
            match node:
                case InputNode(name, types=types):
                    if types is None:
                        types = set[type]()
                    clone.create_input(name, *types)
                case LiteralNode(name, value):
                    clone._nodes[name] = LiteralNode(name, value)
                case ComponentConstructorNode(name, comp, config):
                    cn = clone.add_component(name, comp, config)
                case ComponentInstanceNode(name, comp):
                    cn = clone.add_component(name, comp)
                case _:  # pragma: nocover
                    raise RuntimeError(f"invalid node {node}")

        for n, t in self._aliases.items():
            clone.alias(n, t.name)

        for node in self.nodes():
            match node:
                case ComponentNode(name):
                    wiring = self._edges.get(name, {})
                    cn = clone.node(name)
                    clone.connect(cn, **{wn: clone.node(wt) for (wn, wt) in wiring.items()})
                case _:
                    pass

        for n, t in self._default_connections.items():
            clone.default_connection(n, clone.node(t))

        clone._default = self._default

        return clone

    def build_config(self, *, include_hash: bool = True) -> PipelineConfig:
        """
        Get this pipeline's configuration for serialization.  The configuration
        consists of all inputs and components along with their configurations
        and input connections.  It can be serialized to disk (in JSON, YAML, or
        a similar format) to save a pipeline.

        The configuration does **not** include any trained parameter values,
        although the configuration may include things such as paths to
        checkpoints to load such parameters, depending on the design of the
        components in the pipeline.

        .. note::
            Literal nodes (from :meth:`literal`, or literal values wired to
            inputs) cannot be serialized, and this method will fail if they
            are present in the pipeline.
        """
        self.validate()

        meta = self.meta(include_hash=False)
        cfg = PipelineConfig(meta=meta)

        edges = deepcopy(self._edges)
        for node in self._nodes.values():
            if isinstance(node, ComponentNode):
                c_ins = edges.get(node.name, None)
                if c_ins is None:
                    edges[node.name] = c_ins = {}
                for iname in node.inputs.keys():
                    if iname not in c_ins and iname in self._default_connections:
                        c_ins[iname] = self._default_connections[iname]

        # Now we go over all named nodes and add them to the config:
        for node in self.nodes():
            match node:
                case InputNode():
                    cfg.inputs.append(config.PipelineInput.from_node(node))
                case LiteralNode(name, value):
                    cfg.literals[name] = config.PipelineLiteral.represent(value)
                case ComponentNode(name):
                    c_cfg = config.PipelineComponent.from_node(node)
                    c_cfg.inputs = dict(sorted(edges.get(name, {}).items(), key=lambda kv: kv[0]))
                    cfg.components[name] = c_cfg
                case _:  # pragma: nocover
                    raise RuntimeError(f"invalid node {node}")

        cfg.aliases = {a: t.name for (a, t) in sorted(self._aliases.items(), key=lambda kv: kv[0])}

        if self._default:
            cfg.default = self._default

        cfg.hooks.run = {}
        for name, hooks in self._run_hooks.items():
            cfg.hooks.run[name] = [PipelineHook.from_entry(e) for e in hooks]

        if include_hash:
            cfg.meta.hash = config.hash_config(cfg)

        return cfg

    def config_hash(self) -> str:
        """
        Get a hash of the pipeline's configuration to uniquely identify it for
        logging, version control, or other purposes.

        The hash format and algorithm are not guaranteed, but hashes are stable
        within a LensKit version.  For the same version of LensKit and component
        code, the same configuration will produce the same hash, so long as
        there are no literal nodes.  Literal nodes will *usually* hash
        consistently, but since literals other than basic JSON values are hashed
        by pickling, hash stability depends on the stability of the pickle
        bytestream.

        In LensKit 2025.1, the configuration hash is computed by computing the
        JSON serialization of the pipeline configuration *without* a hash and
        returning the hex-encoded SHA256 hash of that configuration.
        """

        cfg = self.build_config(include_hash=False)
        return config.hash_config(cfg)

    @classmethod
    def load_config(cls, cfg_file: Path | PathLike[str]) -> PipelineBuilder:
        """
        Load a pipeline from a saved configuration file.

        Args:
            cfg_file:
                The path to a TOML, YAML, or JSON file containing the pipeline
                configuration.

        Returns:
            The consructed pipeline.

        See Also:
            :meth:`from_config` for the actual pipeline instantiation logic.
        """
        cfg_file = Path(cfg_file)
        data = load_config_data(cfg_file, PipelineConfig)
        return cls.from_config(data, file_path=cfg_file)

    @classmethod
    def from_config(cls, config: object, *, file_path: Path | None = None) -> PipelineBuilder:
        """
        Reconstruct a pipeline builder from a serialized configuration.

        Args:
            config:
                The configuration object, as loaded from JSON, TOML, YAML, or
                similar. Will be validated into a :class:`PipelineConfig`.
        Returns:
            The configured (but not trained) pipeline.
        Raises:
            PipelineError:
                If there is a configuration error reconstructing the pipeline.
        Warns:
            PipelineWarning:
                If the configuration is funny but usable; for example, the
                configuration includes a hash but the constructed pipeline does
                not have a matching hash.
        """
        from .common import topn_builder, topn_predict_builder

        cfg = PipelineConfig.model_validate(config)

        base = cfg.options.base if cfg.options else None
        if base is None:
            extend = False
            builder = cls()
        else:
            extend = True
            match base:
                case "std:topn":
                    builder = topn_builder(cfg.meta.name, cfg.options)
                case "std:topn-predict":
                    builder = topn_predict_builder(cfg.meta.name, cfg.options)
                case s if s.startswith("std:"):
                    raise ValueError(f"unknown pipeline base {s}")
                case _:
                    raise NotImplementedError("non-standard bases not yet supported")

        builder.apply_config(cfg, extend=extend)

        return builder

    def apply_config(
        self, config: PipelineConfig, *, extend: bool = False, check_hash: bool = False
    ):
        """
        Apply a configuration to this builder.

        Args:
            config:
                The pipeline configuration to apply.
            extend:
                Whether the configuration should extend the current pipeline, or
                fail when there are conflicting definitions.
        """
        self.name = config.meta.name
        self.version = config.meta.version

        for inpt in config.inputs:
            types: list[type[Any] | None] = []
            if inpt.types is not None:
                types += [resolve_type_string(t) for t in inpt.types]
            self.create_input(inpt.name, *types)

        # we now add the components and other nodes in multiple passes to ensure
        # that nodes are available before they are wired (since `connect` can
        # introduce out-of-order dependencies).

        # pass 1: add literals
        for name, data in config.literals.items():
            self.literal(data.decode(), name=name)

        # pass 2: add components
        to_wire: list[config.PipelineComponent] = []
        for name, comp in config.components.items():
            if comp.code.startswith("@"):
                # ignore special nodes in first pass
                continue

            ctor = resolve_type_string(comp.code)
            if isinstance(ctor, type) and issubclass(ctor, Component):
                cfg = ctor.validate_config(comp.config)
            else:
                cfg = None

            if extend and name in self._aliases:
                del self._aliases[name]
            if extend and name in self._nodes:
                self.replace_component(name, ctor, cfg)
            else:
                self.add_component(name, ctor, cfg)
            to_wire.append(comp)

        # pass 3: wiring
        for name, comp in config.components.items():
            inputs = {n: self.node(t) for (n, t) in comp.inputs.items()}
            self.connect(name, **inputs)

        # pass 4: aliases
        for n, t in config.aliases.items():
            self.alias(n, t, replace=extend)

        self._default = config.default
        self._run_hooks = {
            name: [HookEntry(resolve_type_string(h.function), h.priority) for h in hooks]
            for name, hooks in config.hooks.run.items()
        }

        if check_hash and config.meta.hash is not None:
            h2 = self.config_hash()
            if h2 != config.meta.hash:
                _log.warning("loaded pipeline does not match hash")
                warnings.warn(
                    "loaded pipeline config does not match hash", PipelineWarning, stacklevel=2
                )

    def use_first_of(self, name: str, primary: Node[T | None], fallback: Node[T]) -> Node[T]:
        """
        Ergonomic method to create a new node that returns the result of its
        ``input`` if it is provided and not ``None``, and otherwise returns the
        result of ``fallback``.  This method is used for things like filling in
        optional pipeline inputs.  For example, if you want the pipeline to take
        candidate items through an ``items`` input, but look them up from the
        user's history and the training data if ``items`` is not supplied, you
        would do:

        .. code:: python

            pipe = Pipeline()
            # allow candidate items to be optionally specified
            items = pipe.create_input('items', list[EntityId], None)
            # find candidates from the training data (optional)
            lookup_candidates = pipe.add_component(
                'select-candidates', UnratedTrainingItemsCandidateSelector(),
                user=history,
            )
            # if the client provided items as a pipeline input, use those; otherwise
            # use the candidate selector we just configured.
            candidates = pipe.use_first_of('candidates', items, lookup_candidates)

        .. note::

            This method does not distinguish between an input being unspecified
            and explicitly specified as ``None``.

        .. note::

            This method does *not* implement item-level fallbacks, only
            fallbacks at the level of entire results.  For item-level score
            fallbacks, see :class:`~lenskit.basic.FallbackScorer`.

        .. note::
            If one of the fallback elements is a component ``A`` that depends on
            another component or input ``B``, and ``B`` is missing or returns
            ``None`` such that ``A`` would usually fail, then ``A`` will be
            skipped and the fallback will move on to the next node. This works
            with arbitrarily-deep transitive chains.

        Args:
            name:
                The name of the node.
            primary:
                The node to use as the primary input, if it is available.
            fallback:
                The node to use if the primary input does not provide a value.
        """
        return self.add_component(name, fallback_on_none, primary=primary, fallback=fallback)

    def build(self, cache: PipelineCache | None = None) -> Pipeline:
        """
        Build the pipeline.

        Args:
            cache:
                The pipeline cache to use.
        """
        config = self.build_config()
        return Pipeline(
            config,
            [self._instantiate(n, cache) for n in self._nodes.values()],
            run_hooks=cast(RunHooks, self._run_hooks),
        )

    def _instantiate(self, node: Node[ND], cache: PipelineCache | None = None) -> Node[ND]:
        match node:
            case ComponentConstructorNode(name, constructor, config):
                if cache is None:
                    _log.debug("instantiating component", component=constructor)
                    if isinstance(constructor, type) and issubclass(constructor, Component):
                        instance = constructor(config)
                    else:
                        instance = constructor()
                else:
                    instance = cache.get_instance(constructor, config)
                return ComponentInstanceNode(name, instance)
            case _:
                return node

    def _check_available_name(self, name: str) -> None:
        if name in self._nodes or name in self._aliases:
            raise ValueError(f"pipeline already has node {name}")

    def _check_member_node(self, node: Node[Any]) -> None:
        nw = self._nodes.get(node.name)
        if nw is not node:
            raise PipelineError(f"node {node} not in pipeline")



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import hypothesis.strategies as st
from hypothesis import assume, given

from lenskit.data.types import Extent


@given(st.integers(0), st.integers(0))
def test_extent_size(lb: int, ub: int):
    assume(ub >= lb)

    extent = Extent(lb, ub)
    assert extent.start == lb
    assert extent.end == ub
    assert extent.size == ub - lb



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Tests for the Dataset class.
"""

import numpy as np
import pandas as pd

from pytest import approx

from lenskit.data import Dataset
from lenskit.testing import ml_ds, ml_ratings  # noqa: F401


def test_item_stats(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    stats = ml_ds.item_stats()
    stats.info()

    assert len(stats) == ml_ds.item_count
    assert np.all(stats.index == ml_ds.items.index)
    assert stats.index.name == "item_id"

    assert np.all(
        stats["count"] == ml_ratings["item_id"].value_counts().reindex(ml_ds.items, fill_value=0)
    )
    assert np.all(
        stats["user_count"]
        == ml_ratings["item_id"].value_counts().reindex(ml_ds.items, fill_value=0)
    )
    assert np.all(
        stats["rating_count"]
        == ml_ratings["item_id"].value_counts().reindex(ml_ds.items, fill_value=0)
    )

    assert stats["mean_rating"].values == approx(
        ml_ratings.groupby("item_id")["rating"].mean().reindex(ml_ds.items).values, nan_ok=True
    )

    ts = ml_ratings.groupby("item_id")["timestamp"].min().reindex(ml_ds.items)
    bad = stats["first_time"] != ts
    bad &= stats["first_time"].isnull() != ts.isnull()
    nbad = np.sum(bad)
    if nbad:
        df = stats[["first_time"]].assign(expected=ts)
        bdf = df[bad]
        raise AssertionError(f"timestamps mismatch:\n{bdf}")


def test_user_stats(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    stats = ml_ds.user_stats()
    stats.info()

    assert len(stats) == ml_ds.user_count
    assert np.all(stats.index == ml_ds.users.index)
    assert stats.index.name == "user_id"

    assert np.all(stats["count"] == ml_ratings["user_id"].value_counts().reindex(ml_ds.users))
    assert np.all(stats["item_count"] == ml_ratings["user_id"].value_counts().reindex(ml_ds.users))
    assert np.all(
        stats["rating_count"] == ml_ratings["user_id"].value_counts().reindex(ml_ds.users)
    )

    assert stats["mean_rating"].values == approx(
        ml_ratings.groupby("user_id")["rating"].mean().reindex(ml_ds.users).values
    )
    assert np.all(
        stats["first_time"] == ml_ratings.groupby("user_id")["timestamp"].min().reindex(ml_ds.users)
    )
    assert np.all(
        stats["last_time"] == ml_ratings.groupby("user_id")["timestamp"].max().reindex(ml_ds.users)
    )



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
import warnings
from typing import TypeVar

import numpy as np
import pandas as pd

from lenskit.data import ItemList, ItemListCollection
from lenskit.diagnostics import DataWarning
from lenskit.logging import item_progress
from lenskit.metrics._accum import MetricAccumulator

from ._base import Metric, MetricFunction

_log = logging.getLogger(__name__)
K1 = TypeVar("K1", bound=tuple)
K2 = TypeVar("K2", bound=tuple)


class RunAnalysisResult:
    """
    Results of a bulk metric computation.

    Stability:
        Caller
    """

    _list_metrics: pd.DataFrame
    _global_metrics: pd.Series
    _defaults: dict[str, float]

    def __init__(self, lmvs: pd.DataFrame, gmvs: pd.Series, defaults: dict[str, float]):
        self._list_metrics = lmvs
        self._global_metrics = gmvs
        self._defaults = defaults

    def global_metrics(self) -> pd.Series:
        """
        Get the global metric scores.  This is only the results of
        global metrics; it does not include aggregates of per-list metrics.  For
        aggregates of per-list metrics, call :meth:`list_summary`.
        """
        return self._global_metrics

    def list_metrics(self, fill_missing=True) -> pd.DataFrame:
        """
        Get the per-list scores of the results.  This is a data frame with one
        row per list (with the list key on the index), and one metric per
        column.

        Args:
            fill_missing:
                If ``True`` (the default), fills in missing values with each
                metric's default value when available.  Pass ``False`` if you
                want to do analyses that need to treat missing values
                differently.
        """
        if fill_missing:
            return self._list_metrics.fillna(self._defaults)
        return self._list_metrics

    def list_summary(self, *keys: str) -> pd.DataFrame:
        """
        Summary statistics for the per-list metrics.  Each metric is on its own
        row, with columns reporting the following:

        ``mean``:
            The mean metric value.
        ``median``:
            The median metric value.
        ``std``:
            The (sample) standard deviation of the metric.

        Additional columns are added based on other options.  Missing metric
        values are filled with their defaults before computing statistics.

        Args:
            keys:
                Identifiers for different conditions that should be reported
                separately (grouping keys for the final result).
        """
        scores = self.list_metrics(fill_missing=True)
        if keys:
            df = scores.groupby(list(keys)).agg(["mean", "median", "std"]).stack(level=0)
            assert isinstance(df, pd.DataFrame)
        else:
            df = scores.agg(["mean", "median", "std"]).T
            df.index.name = "metric"
        return df

    def merge_from(self, other: RunAnalysisResult):
        """
        Merge another set of analysis results into this one.
        """
        for c in self._list_metrics.columns:
            if c in other._list_metrics.columns:
                warnings.warn(f"list metric {c} appears in both merged results", DataWarning)

        for c in self._global_metrics.index:
            if c in other._global_metrics.index:
                warnings.warn(f"global metric {c} appears in both merged results", DataWarning)

        self._list_metrics = self._list_metrics.join(other._list_metrics, how="outer")
        self._global_metrics = pd.concat([self._global_metrics, other._global_metrics])


class RunAnalysis:
    """
    Compute metrics over a collection of item lists composing a run.

    This class now uses :class:`MetricAccumulator` internally to separate
    accumulation from looping, while maintaining the same external interface.

    Args:
        metrics:
            A list of metrics; you can also add them with :meth:`add_metric`,
            which provides more flexibility.

    Stability:
        Caller
    """

    def __init__(self, *metrics: Metric):
        self._accumulator = MetricAccumulator()
        for metric in metrics:
            self._accumulator.add_metric(metric)

    def add_metric(
        self,
        metric: Metric | MetricFunction | type[Metric],
        label: str | None = None,
        default: float | None = None,
    ):
        """
        Add a metric to this metric set.

        Args:
            metric:
                The metric to add to the set.
            label:
                The label to use for the metric's results.  If unset, obtains
                from the metric.
            default:
                The default value to use in aggregates when a user does not have
                recommendations. If unset, obtains from the metric's ``default``
                attribute (if specified).
        """
        self._accumulator.add_metric(metric, label, default)

    @property
    def metrics(self) -> list:
        """
        The list of metrics to compute.

        .. deprecated:: 2025.4
            Access metrics through the accumulator interface instead.
        """
        return self._accumulator.metrics

    def compute(
        self, outputs: ItemListCollection[K1], test: ItemListCollection[K2]
    ) -> RunAnalysisResult:
        """
        Deprecated alias for :meth:`measure`.

        .. deprecated:: 2025.1.1
            Use :meth:`measure` instead.
        """
        return self.measure(outputs, test)

    def measure(
        self, outputs: ItemListCollection[K1], test: ItemListCollection[K2]
    ) -> RunAnalysisResult:
        """
        Measure a set of outputs against a set of test data.
        """
        self._accumulator._validate_setup()

        if len(outputs.key_fields) > 1:
            index = pd.MultiIndex.from_tuples(outputs.keys())
            index.names = list(outputs.key_fields)
        else:
            index = pd.Index([k[0] for k in outputs.keys()])
            index.name = outputs.key_fields[0]

        n = len(outputs)
        _log.debug("measuring %d metrics for %d output lists", len(self._accumulator.metrics), n)

        no_test_count = 0
        with item_progress("Measuring", n) as pb:
            for key, out in outputs:
                key_kwargs = dict(zip(outputs.key_fields, key))
                list_test = test.lookup_projected(key)
                if out is None or (list_test is None and not out):
                    continue
                elif list_test is None:
                    no_test_count += 1
                    list_test = ItemList([])
                self._accumulator.measure_list(out, list_test, **key_kwargs)
                pb.update()

        if no_test_count:
            _log.warning("could not find test data for %d lists", no_test_count)

        list_results = self._accumulator.list_metrics(fill_missing=False)

        global_metrics = [m for m in self._accumulator.metrics if m.is_global]

        _log.debug("computing %d global metrics", len(global_metrics))

        global_results = {}
        for metric_wrapper in global_metrics:
            result = metric_wrapper.measure_run(outputs, test)

            if result is None and metric_wrapper.default is not None:
                result = metric_wrapper.default

            if result is not None:
                global_results[metric_wrapper.label] = result

        summary_results = self._accumulator.summary_metrics()
        for key, value in summary_results.items():
            if "." in key:
                base_metric = key.split(".", 1)[0]
                if base_metric not in global_results:
                    global_results[base_metric] = value
            else:
                if key not in global_results:
                    global_results[key] = value

        global_results = pd.Series(global_results, dtype=np.float64)

        defaults = {
            metric_wrapper.label: metric_wrapper.default
            for metric_wrapper in self._accumulator.metrics
            if metric_wrapper.default is not None
        }

        return RunAnalysisResult(list_results, global_results, defaults)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import pandas as pd  # noqa: 401

from pytest import approx, raises

from lenskit.data import ItemListCollection
from lenskit.data.adapt import ITEM_COMPAT_COLUMN, USER_COMPAT_COLUMN
from lenskit.metrics._base import GlobalMetric
from lenskit.metrics.basic import ListLength
from lenskit.metrics.bulk import RunAnalysis
from lenskit.metrics.predict import RMSE
from lenskit.metrics.ranking import DCG, NDCG, RBP, Precision, RecipRank
from lenskit.testing import demo_recs, ml_ratings


def test_bulk_measure_function(ml_ratings: pd.DataFrame):
    bms = RunAnalysis()
    bms.add_metric(ListLength(), "length")
    bms.add_metric(RMSE)

    class DummyGlobalMetric(GlobalMetric):
        def measure_run(self, output, test):
            return None

    bms.add_metric(DummyGlobalMetric(), label="global_metric", default=123.0)

    data = ItemListCollection.from_df(
        ml_ratings.rename(columns={"rating": "score"}), USER_COMPAT_COLUMN, ITEM_COMPAT_COLUMN
    )
    truth = ItemListCollection.from_df(ml_ratings, USER_COMPAT_COLUMN, ITEM_COMPAT_COLUMN)

    metrics = bms.measure(data, truth)

    # check normal metrics
    stats = metrics.list_summary()
    assert stats.loc["length", "mean"] == approx(ml_ratings["user_id"].value_counts().mean())
    assert stats.loc["RMSE", "mean"] == approx(0)

    # check global fallback metric
    global_metrics = metrics.global_metrics()
    assert global_metrics["global_metric"] == 123.0


def test_recs(demo_recs):
    split, recs = demo_recs

    bms = RunAnalysis()
    bms.add_metric(ListLength())
    bms.add_metric(Precision())
    bms.add_metric(NDCG())
    bms.add_metric(DCG())
    bms.add_metric(RBP)
    bms.add_metric(RecipRank)

    metrics = bms.measure(recs, split.test)
    scores = metrics.list_metrics()
    stats = metrics.list_summary()
    print(stats)
    for m in bms.metrics:
        assert stats.loc[m.label, "mean"] == approx(scores[m.label].mean())


def test_duplicate_metric(demo_recs):
    split, recs = demo_recs

    bms = RunAnalysis()
    bms.add_metric(NDCG(k=10))
    bms.add_metric(NDCG(k=10))
    bms.add_metric(RBP)
    bms.add_metric(RecipRank)

    with raises(RuntimeError, match=r"duplicate"):
        bms.measure(recs, split.test)


def test_recs_multi(demo_recs):
    split, recs = demo_recs

    il2 = ItemListCollection(["rep", "user_id"])
    il2.add_from(recs, rep=1)
    il2.add_from(recs, rep=2)

    bms = RunAnalysis()
    bms.add_metric(ListLength())
    bms.add_metric(Precision())
    bms.add_metric(NDCG())
    bms.add_metric(RBP)
    bms.add_metric(RecipRank)

    metrics = bms.measure(il2, split.test)
    scores = metrics.list_metrics()
    stats = metrics.list_summary("rep")
    print(stats)
    for m in bms.metrics:
        assert stats.loc[(1, m.label), "mean"] == approx(scores.loc[1, m.label].mean())
        assert stats.loc[(2, m.label), "mean"] == approx(scores.loc[2, m.label].mean())



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle
from itertools import groupby

import numpy as np
import pandas as pd
from numpy.typing import NDArray

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
from hypothesis import HealthCheck, assume, given, settings

from lenskit.basic.random import RandomSelector
from lenskit.data.items import ItemList
from lenskit.testing import scored_lists

_log = logging.getLogger(__name__)


@given(scored_lists())
def test_unlimited_selection(items: ItemList):
    rsel = RandomSelector()
    ranked = rsel(items=items)

    ids = items.ids()

    assert len(ranked) == len(ids)
    assert set(ranked.ids()) == set(ids)


@given(st.integers(min_value=1, max_value=100), scored_lists())
def test_configured_truncation(n, items: ItemList):
    rsel = RandomSelector(n=n)
    ranked = rsel(items=items)

    assert len(ranked) == min(n, len(items))
    # all items are unique
    assert len(set(ranked.ids())) == len(ranked)


@given(st.integers(min_value=1, max_value=100), scored_lists())
def test_runtime_truncation(n, items: ItemList):
    rsel = RandomSelector()
    ranked = rsel(items=items, n=n)

    assert len(ranked) == min(n, len(items))
    # all items are unique
    assert len(set(ranked.ids())) == len(ranked)


@settings(suppress_health_check=[HealthCheck.too_slow])
@given(
    st.integers(min_value=5, max_value=100),
    st.lists(st.emails(), min_size=5, unique=True),
    scored_lists(n=(5, 500)),
)
def test_random_seeding(n: int, users: list[str], items: ItemList):
    assume(n < len(items))
    rsel = RandomSelector(rng=(42, "user"))

    distinct = set()

    for user in users:
        ranked = rsel(items=items, n=n, query=user)

        assert len(ranked) == n
        # all items are unique
        assert len(set(ranked.ids())) == len(ranked)

        # ranking again gets the same list
        r2 = rsel(items=items, n=n, query=user)
        assert np.all(r2.ids() == ranked.ids())

        distinct |= set(ranked.ids())

    # we should have more than n distinct values (randomness works)
    assert len(distinct) > n



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import numpy as np

from lenskit.data import ItemList


def least_item_promoted(reference: ItemList, reranked: ItemList, k: int = 10) -> float:
    """
    Compute the Least Item Promoted (LIP) metric.

    This metric identifies the item in the top-k reranked list that had the
    highest (worst) rank in the reference, and returns how many positions
    it was promoted from beyond k.

    Args:
        reference:
            The original/base ranking.
        reranked:
            The reranked list.
        k:
            The depth to evaluate (default 10).

    Returns:
        The rank distance of the least-promoted item, or NaN if the base
        ranking is empty.

    Stability:
        Experimental
    """
    if len(reference) == 0:
        return np.nan

    reference_ids = reference.ids()
    lip_rank = k
    for item_id in reranked[:k].ids():
        indices = np.where(reference_ids == item_id)[0]
        if indices.size > 0:
            rank = indices[0]
            if rank > lip_rank:
                lip_rank = rank

    return lip_rank - k



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Support code for managing model state.

Many LensKit components are built on machine learning models of various forms.
While :mod:`lenskit.training` provides support for training those models,
this module provides support for managing their learned state and weights.
"""

from ._container import ParameterContainer

__all__ = [
    "ParameterContainer",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from uuid import uuid4

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import torch

from pytest import mark

from lenskit._accel import data
from lenskit.parallel import ensure_parallel_init
from lenskit.stats import argtopn


@mark.benchmark(group="all")
@mark.parametrize("size", [100, 5000, 100_000, 1_000_000])
def test_numpy_argsort(rng: np.random.Generator, size: int, benchmark):
    scores = rng.standard_exponential(size)
    scores = np.require(scores, dtype=np.float32)

    def sort():
        _idx = np.argsort(-scores)

    benchmark(sort)


@mark.benchmark(group="all")
@mark.parametrize("size", [100, 5000, 100_000, 1_000_000])
def test_torch_argsort(rng: np.random.Generator, size: int, benchmark):
    scores = rng.standard_exponential(size)
    scores = torch.as_tensor(np.require(scores, dtype=np.float32))

    def sort():
        _idx = torch.argsort(-scores)

    benchmark(sort)


@mark.benchmark(group="all")
@mark.parametrize("size", [100, 5000, 100_000, 1_000_000])
@mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_torch_argsort_gpu(rng: np.random.Generator, size: int, benchmark):
    scores = rng.standard_exponential(size)
    scores = torch.as_tensor(np.require(scores, dtype=np.float32)).cuda()

    def sort():
        _idx = torch.argsort(-scores).cpu()

    benchmark(sort)


@mark.benchmark(group="all")
@mark.parametrize("size", [100, 5000, 100_000, 1_000_000])
@mark.skip("only testing argsort")
def test_numpy_sort(rng: np.random.Generator, size: int, benchmark):
    scores = rng.standard_exponential(size)
    scores = np.require(scores, dtype=np.float32)

    def sort():
        _res = np.sort(-scores)

    benchmark(sort)


@mark.benchmark(group="all")
@mark.parametrize("size", [100, 5000, 100_000, 1_000_000])
def test_arrow_sort(rng: np.random.Generator, size: int, benchmark):
    scores = rng.standard_exponential(size)
    scores = pa.array(scores, pa.float32())

    def sort():
        _idx = pc.array_sort_indices(scores, order="descending")

    benchmark(sort)


@mark.benchmark(group="all")
@mark.parametrize("size", [100, 5000, 100_000, 1_000_000])
def test_accel_sort(rng: np.random.Generator, size: int, benchmark):
    ensure_parallel_init()
    scores = rng.standard_exponential(size)
    scores = pa.array(scores, pa.float32())

    def sort():
        _idx = data.argsort_descending(scores)

    benchmark(sort)


@mark.benchmark(group="topn")
@mark.parametrize("size", [100, 5000, 100_000, 1_000_000])
def test_argtopn(rng: np.random.Generator, size: int, benchmark):
    scores = rng.standard_exponential(size)

    def sort():
        _res = argtopn(scores, 100)

    benchmark(sort)


@mark.benchmark(group="topn")
@mark.parametrize("size", [100, 5000, 100_000, 1_000_000])
def test_torch_topk(rng: np.random.Generator, size: int, benchmark):
    scores = rng.standard_exponential(size)
    scores = torch.as_tensor(scores)

    def sort():
        _idx = torch.topk(-scores, 100)

    benchmark(sort)


@mark.benchmark(group="topn")
@mark.parametrize("size", [100, 5000, 100_000, 1_000_000])
@mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_torch_topk_gpu(rng: np.random.Generator, size: int, benchmark):
    scores = rng.standard_exponential(size)
    scores = torch.as_tensor(scores).cuda()

    def sort():
        _s, _idx = torch.topk(-scores, 100)
        _s = _s.cpu()
        _idx = _idx.cpu()

    benchmark(sort)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from time import sleep

from pytest import importorskip, mark

from lenskit.logging import get_logger
from lenskit.parallel.ray import TaskLimiter, ensure_cluster, ray_available

ray = importorskip("ray")
pytestmark = mark.skipif(not ray_available(), reason="ray is not available")

_log = get_logger(__name__)


@ray.remote
def _dummy_task(n):
    _log.info("worker task %d", n)
    sleep(0.1)
    return n * 100


def test_task_limiter():
    ensure_cluster()
    NTASK = 4
    limit = TaskLimiter(NTASK)

    n_pend = []

    for i in range(20):
        n_pend.append(limit.pending)
        assert limit.pending <= NTASK
        limit.wait_for_limit()

        task = _dummy_task.remote(i)
        limit.add_task(task)

    assert limit.pending <= NTASK
    limit.drain()
    assert limit.pending == 0

    # make sure we were parallelized
    assert any(n >= 2 for n in n_pend)


def test_map():
    ensure_cluster()
    NTASK = 4
    limit = TaskLimiter(NTASK)

    for i, result in enumerate(limit.imap(_dummy_task, range(50))):
        assert result == i * 100


def test_map_unordered():
    ensure_cluster()
    NTASK = 4
    limit = TaskLimiter(NTASK)

    results = set()

    for result in limit.imap(_dummy_task, range(50), ordered=False):
        results.add(result)

    assert results == set(i * 100 for i in range(50))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Simple (non-iterative) evaluation of points.
"""

from __future__ import annotations

from pydantic_core import to_json

from lenskit.logging import Task, get_logger
from lenskit.logging.worker import send_task
from lenskit.pipeline import Pipeline
from lenskit.random import make_seed
from lenskit.training import TrainingOptions

from ._job import TuningJobData
from ._measure import measure_pipeline

_log = get_logger(__name__)


class SimplePointEval:
    """
    A simple hyperparameter point evaluator using non-iterative model training.
    """

    job: TuningJobData

    def __init__(self, job: TuningJobData):
        self.job = job

    def __call__(self, config) -> dict[str, float]:
        comp_name = self.job.spec.component_name
        assert comp_name is not None
        cfg = self.job.pipeline.merge_component_configs({comp_name: config})

        pipe = Pipeline.from_config(cfg)

        data = self.job.data

        rng = make_seed(self.job.random_seed, to_json(config))
        with Task(
            label=f"train {pipe.name}",
            tags=["tune", "train"],
            reset_hwm=True,
            subprocess=True,
        ) as train_task:
            pipe.train(data.train, TrainingOptions(rng=rng))
        send_task(train_task)

        with Task(
            label=f"measure {pipe.name}",
            tags=["tune", "recommend"],
            reset_hwm=True,
            subprocess=True,
        ) as test_task:
            results = measure_pipeline(self.job.spec, pipe, data.test, train_task, test_task)
        send_task(test_task)

        return results



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path

import click
import pandas as pd
from rich.table import Table

from lenskit.logging import stdout_console


@click.command("summarize")
@click.argument("FILE", type=Path, required=True)
def summarize_profile(file: Path):
    """
    Summarize a pipeline profile.
    """

    console = stdout_console()
    prof = pd.read_csv(file)
    # convert to ms
    prof = prof / 1_000_000
    summary = prof.agg(["mean", "std", "median"]).T
    summary["95%ile"] = prof.quantile(0.95)
    summary.index.name = "Stage"

    tbl = Table(title="Stage Timing (ms)")
    tbl.add_column("Stage", justify="left")
    for c in summary.columns:
        tbl.add_column(c, justify="right")

    for row in summary.itertuples():
        tbl.add_row(row[0], *["{:.4f}".format(r) for r in row[1:]])

    console.print(tbl)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Input type checking hook.
"""

from typing import Any

from ..nodes import ComponentInstanceNode
from ..types import SkipComponent, TypeExpr, is_compatible_data


def typecheck_input_data(
    node: ComponentInstanceNode[Any],
    input_name: str,
    input_type: TypeExpr | None,
    value: Any,
    *,
    required: bool = True,
    **context,
) -> Any:
    """
    Hook to check that input data matches the component's declared input type.
    """
    if input_type and not is_compatible_data(value, input_type):
        if value is None:
            bad_type = None
        else:
            bad_type = type(value)
        msg = f"found {bad_type}, expected ❬{input_type}❭"
        if value is None and not required:
            err = SkipComponent(msg)
        else:
            err = TypeError(msg)
        err.add_note(f"Error encountered on input ❬{input_name}❭ to component ❬{node.name}❭")
        raise err

    return value



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd
import pyarrow as pa

from lenskit._accel import data
from lenskit.data.vocab import Vocabulary


def test_empty():
    tbl = data.CoordinateTable(5)
    assert len(tbl) == 0
    assert tbl.dimensions() == 5
    assert tbl.unique_count() == 0


def test_add_rb():
    tbl = data.CoordinateTable(2)
    batch = pa.record_batch(
        {
            "row_num": pa.array([0, 0, 1, 1, 1, 3, 3], pa.int32()),
            "col_num": pa.array([0, 1, 0, 2, 5, 1, 3], pa.int32()),
        }
    )

    n, nuq = tbl.extend(batch)
    assert n == 7
    assert nuq == 7

    assert tbl.dimensions() == 2
    assert len(tbl) == 7
    assert tbl.unique_count() == 7

    assert tbl.contains(0, 1)
    assert not tbl.contains(0, 5)

    assert tbl.find(1, 2) == 3
    assert tbl.find(1, 3) is None


def test_add_dupes():
    tbl = data.CoordinateTable(2)
    batch = pa.record_batch(
        {
            "row_num": pa.array([0, 0, 1, 1, 1, 3, 3, 1], pa.int32()),
            "col_num": pa.array([0, 1, 0, 2, 5, 1, 3, 2], pa.int32()),
        }
    )

    n, nuq = tbl.extend(batch)
    assert n == 8
    assert nuq == 7

    assert tbl.dimensions() == 2
    assert len(tbl) == 8
    assert tbl.unique_count() == 7

    assert tbl.contains(0, 1)
    assert tbl.contains(1, 2)

    assert tbl.find(1, 2) == 7
    assert tbl.find(1, 3) is None


def test_add_batches(ml_ratings: pd.DataFrame, rng: np.random.Generator):
    ml_tbl = pa.Table.from_pandas(ml_ratings, preserve_index=False)

    items = Vocabulary(ml_tbl.column("item_id"), reorder=True)
    users = Vocabulary(ml_tbl.column("user_id"), reorder=True)

    ui_tbl = pa.table(
        {
            "user_num": users.numbers(ml_tbl.column("user_id"), format="arrow"),
            "item_num": items.numbers(ml_tbl.column("item_id"), format="arrow"),
        }
    )

    tbl = data.CoordinateTable(2)
    n, nuq = tbl.extend(ui_tbl.to_batches(10000))

    assert n == ml_tbl.num_rows
    assert nuq == n

    assert tbl.dimensions() == 2
    assert len(tbl) == n
    assert tbl.unique_count() == n

    for pos in rng.choice(ml_tbl.num_rows, size=500):
        uno = tbl._get_coord(0, pos)
        assert uno == ui_tbl.column("user_num")[pos].as_py()

        ino = tbl._get_coord(1, pos)
        assert ino == ui_tbl.column("item_num")[pos].as_py()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
from dataclasses import dataclass

from lenskit.pipeline.components import Component, component_inputs, component_return_type


@dataclass
class XConfig:
    suffix: str = ""


class XComp(Component):
    config: XConfig

    def __call__(self, msg: str) -> str:
        return msg + self.config.suffix


class CallObj:
    def __call__(self, q: str) -> bytes:
        return q.encode()


def test_empty_input():
    def func() -> int:
        return 9

    inputs = component_inputs(func)
    assert not inputs


def test_single_function_input():
    def func(x: int) -> int:
        return 9 + x

    inputs = component_inputs(func)
    assert len(inputs) == 1
    assert inputs["x"] is int


def test_component_class_input():
    inputs = component_inputs(XComp)
    assert len(inputs) == 1
    assert inputs["msg"] is str


def test_component_object_input():
    inputs = component_inputs(XComp())
    assert len(inputs) == 1
    assert inputs["msg"] is str


def test_component_unknown_input():
    def func(x) -> int:  # type: ignore
        return x + 5  # type: ignore

    inputs = component_inputs(func)  # type: ignore
    assert len(inputs) == 1
    assert inputs["x"] is None


def test_callable_object_input():
    inputs = component_inputs(CallObj())
    assert len(inputs) == 1
    assert inputs["q"] is str


def test_callable_class_input():
    inputs = component_inputs(CallObj)
    assert len(inputs) == 1
    assert inputs["q"] is str


def test_function_return():
    def func(x: int) -> int:
        return x + 5

    rt = component_return_type(func)
    assert rt is int


def test_class_return():
    rt = component_return_type(XComp)
    assert rt is str


def test_instance_return():
    rt = component_return_type(XComp())
    assert rt is str


def test_unknown_return():
    def func():
        pass

    rt = component_return_type(func)
    assert rt is None


def test_callable_object_return():
    rt = component_return_type(CallObj())
    assert rt is bytes


def test_callable_class_return():
    rt = component_return_type(CallObj)
    assert rt is bytes



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from uuid import uuid4

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc

from pytest import mark

from lenskit.data import ItemList, Vocabulary

LEFT_SIZE = 50000
RIGHT_SIZE = 100
TOTAL_SIZE = LEFT_SIZE * 2


@mark.benchmark(group="numbers")
@mark.parametrize("size", [100, 5000, 10_000, 100_000, 1_000_000])
def test_itemlist_remove_numbers(rng: np.random.Generator, size: int, benchmark):
    vocab = Vocabulary(np.arange(size * 2))
    left = np.arange(size)
    right = rng.choice(size * 2, RIGHT_SIZE, replace=False)

    il_left = ItemList(item_nums=left, vocabulary=vocab, scores=rng.standard_exponential(size))

    def search():
        _rem = il_left.remove(numbers=right)

    benchmark(search)


@mark.benchmark(group="numbers")
@mark.parametrize("size", [100, 5000, 10_000, 100_000, 1_000_000])
def test_numpy_isin_numbers(rng: np.random.Generator, size: int, benchmark):
    vocab = Vocabulary(np.arange(size * 2))
    left = np.arange(size)
    right = rng.choice(size * 2, RIGHT_SIZE, replace=False)

    il_left = ItemList(item_nums=left, vocabulary=vocab, scores=rng.standard_exponential(size))

    def search():
        mask = np.isin(il_left.numbers(), right)
        _rem = il_left[mask]

    benchmark(search)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Arryw utility functions.
"""

from functools import partial

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import torch
from typing_extensions import Callable, Literal, TypeAlias, TypeVar, overload

from .mtarray import MTArray

A = TypeVar("A", bound=pa.Array, default=pa.Array)
Selector: TypeAlias = Callable[[MTArray | A | None], A | None]


def get_indexer(sel) -> Selector:
    """
    Get a selector that will apply the specified indexer.  This allows
    one indexer to be applied to multiple arrays.
    """
    if np.isscalar(sel):
        sel = pa.array([sel], pa.int32())  # type: ignore
        return partial(arrow_take, sel)
    elif isinstance(sel, slice):
        return partial(arrow_slice, sel)
    elif not isinstance(sel, pa.Array):
        sel = pa.array(sel)

    if pa.types.is_integer(sel.type):
        return partial(arrow_take, sel)
    elif pa.types.is_boolean(sel.type):
        return partial(arrow_filter, sel)
    else:  # pragma: nocover
        raise TypeError(f"invalid selector: {sel}")


def arrow_to_format(array: pa.Array, format: Literal["arrow", "numpy", "torch"]):
    """
    Convert an Arrow array into another format.

    Stability:
        Internal
    """
    match format:
        case "arrow":
            return array
        case "numpy":
            return array.to_numpy(zero_copy_only=False)
        case "torch":
            return torch.as_tensor(array.to_numpy(zero_copy_only=False, writable=True))
        case _:  # pragma: nocover
            raise ValueError(f"unknown format {format}")


@overload
def arrow_slice(sel: slice, array: MTArray | A) -> A: ...
@overload
def arrow_slice(sel: slice, array: MTArray | A | None) -> A | None: ...
def arrow_slice(sel: slice, array: MTArray | A | None) -> A | None:
    """
    Slice an Arrow array.
    """
    if array is None:
        return None
    elif isinstance(array, MTArray):
        array = array.arrow()

    if sel.step and sel.step != 1:
        raise ValueError("slices with steps unsupported")
    start = sel.start or 0
    if sel.stop is None:
        slen = len(array) - start
    elif sel.stop < 0:
        slen = len(array) - start + sel.stop
    else:
        slen = sel.stop - start

    return array.slice(start, slen)


@overload
def arrow_take(sel: pa.Int32Array, array: MTArray | A) -> A: ...
@overload
def arrow_take(sel: pa.Int32Array, array: MTArray | A | None) -> A | None: ...
def arrow_take(sel: pa.Int32Array, array: MTArray | A | None) -> A | None:
    """
    Select from an Arrow array by integer indices.
    """

    if array is None:
        return None
    elif isinstance(array, MTArray):
        array = array.arrow()

    return array.take(sel)


@overload
def arrow_filter(sel: pa.BooleanArray, array: MTArray | A) -> A: ...
@overload
def arrow_filter(sel: pa.BooleanArray, array: MTArray | A | None) -> A | None: ...
def arrow_filter(sel: pa.BooleanArray, array: MTArray | A | None) -> A | None:
    """
    Select from an Arrow array by integer indices.
    """

    if array is None:
        return None
    elif isinstance(array, MTArray):
        array = array.arrow()

    return array.filter(sel)


def arrow_type(dtype: np.dtype) -> pa.DataType:
    """
    Resolve a NumPy data type to an Arrow data type, assuming objects store
    strings.
    """
    if dtype == np.object_:
        return pa.utf8()
    else:
        return pa.from_numpy_dtype(dtype)


def explode_column(tbl: pa.Table, column: str, *, out_col: str | None = None) -> pa.Table:
    """
    “Explode” a list column to yield one row per list element.
    """
    col = tbl.column(column)
    if not (
        pa.types.is_list(col.type)
        or pa.types.is_list_view(col.type)
        or pa.types.is_large_list(col.type)
        or pa.types.is_large_list_view(col.type)
    ):
        raise TypeError(f"column type {col.type} is not a list type")

    paired = tbl.drop_columns(column)
    indices = pc.list_parent_indices(col)
    exp_tbl = paired.take(indices)
    exp_col = pc.list_flatten(col)
    assert len(exp_tbl) == len(exp_col)
    return exp_tbl.append_column(out_col or column, exp_col)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
from __future__ import annotations

from abc import abstractmethod
from collections.abc import Mapping
from typing import Any, cast

from pydantic import JsonValue
from typing_extensions import Generic, TypeVar

from .components import (
    Component,
    ComponentConstructor,
    PipelineFunction,
    component_inputs,
    component_return_type,
)

# Nodes are (conceptually) immutable data containers, so Node[U] can be assigned
# to Node[T] if U ≼ T.
ND = TypeVar("ND", covariant=True)
CFG = TypeVar("CFG", contravariant=True, bound=object)


class Node(Generic[ND]):
    """
    Representation of a single node in a :class:`Pipeline`.

    Stability:
        Caller
    """

    __match_args__ = ("name",)

    name: str
    "The name of this node."
    types: set[type] | None
    "The set of valid data types of this node, or None for no typechecking."

    def __init__(self, name: str, *, types: set[type] | None = None):
        self.name = name
        self.types = types

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} {self.name}>"


class InputNode(Node[ND], Generic[ND]):
    """
    An input node.

    Stability:
        Internal
    """


class LiteralNode(Node[ND], Generic[ND]):
    """
    A node storing a literal value.

    Stability:
        Internal
    """

    __match_args__ = ("name", "value")
    value: ND
    "The value associated with this node"

    def __init__(self, name: str, value: ND, *, types: set[type] | None = None):
        super().__init__(name, types=types)
        self.value = value


class ComponentNode(Node[ND], Generic[ND]):
    """
    A node storing a component.  This is an abstract node class; see subclasses
    :class:`ComponentConstructorNode` and `ComponentInstanceNode`.

    Stability:
        Internal
    """

    def __init__(self, name: str):
        super().__init__(name)

    @staticmethod
    def create(
        name: str,
        comp: ComponentConstructor[CFG, ND] | Component[ND] | PipelineFunction[ND],
        config: CFG | Mapping[str, JsonValue] | None = None,
    ) -> ComponentNode[ND]:
        if isinstance(comp, Component):
            return ComponentInstanceNode(name, cast(Component[ND], comp))
        elif isinstance(comp, ComponentConstructor):
            comp = cast(ComponentConstructor[CFG, ND], comp)
            return ComponentConstructorNode(name, comp, comp.validate_config(config))
        elif isinstance(comp, type):
            return ComponentConstructorNode(name, comp, None)  # type: ignore
        else:
            return ComponentInstanceNode(name, comp)

    @property
    @abstractmethod
    def inputs(self) -> dict[str, type | None]:  # pragma: nocover
        raise NotImplementedError()


class ComponentConstructorNode(ComponentNode[ND], Generic[ND]):
    __match_args__ = ("name", "constructor", "config")
    constructor: ComponentConstructor[Any, ND]
    config: object | None

    def __init__(self, name: str, constructor: ComponentConstructor[CFG, ND], config: CFG | None):
        super().__init__(name)
        self.constructor = constructor
        self.config = config
        if rt := component_return_type(constructor):
            self.types = {rt}

    @property
    def inputs(self):
        return component_inputs(self.constructor)


class ComponentInstanceNode(ComponentNode[ND], Generic[ND]):
    __match_args__ = ("name", "component")

    component: Component[ND] | PipelineFunction[ND]

    def __init__(
        self,
        name: str,
        component: Component[ND] | PipelineFunction[ND],
    ):
        super().__init__(name)
        self.component = component
        if rt := component_return_type(component):
            self.types = {rt}

    @property
    def inputs(self):
        return component_inputs(self.component)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from lenskit import batch
from lenskit.als import BiasedMFScorer
from lenskit.data import Dataset
from lenskit.logging.tasks import Task, TaskStatus
from lenskit.pipeline import topn_pipeline


def test_train_task(ml_ds: Dataset):
    info = BiasedMFScorer(features=50, epochs=5)
    pipe = topn_pipeline(info)

    with Task("train ImplicitMF", reset_hwm=True) as task:
        pipe.train(ml_ds)

    print(task.model_dump_json(indent=2))

    assert task.status == TaskStatus.FINISHED
    assert task.cpu_time is not None and task.cpu_time > 0
    if task.peak_memory is not None:
        assert task.peak_memory > 0

    users = ml_ds.users.ids()[:500]
    with Task("recommend", reset_hwm=True) as task:
        _recs = batch.recommend(pipe, users, n=20, n_jobs=2)  # type: ignore

    print(task.model_dump_json(indent=2))

    assert task.status == TaskStatus.FINISHED
    assert len(task.subtasks) == 2
    assert all(st.subprocess for st in task.subtasks)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path
from shutil import copytree

import numpy as np

from pytest import mark, raises

from lenskit.data.dataset import Dataset
from lenskit.data.movielens import load_movielens, load_movielens_df
from lenskit.testing import ml_test_dir

ML_LATEST_DIR = Path("data/ml-latest-small")

ML_100K_ZIP = Path("data/ml-100k.zip")
ML_1M_ZIP = Path("data/ml-1m.zip")
ML_10M_ZIP = Path("data/ml-10m.zip")
ML_20M_ZIP = Path("data/ml-20m.zip")
ML_25M_ZIP = Path("data/ml-25m.zip")
ML_32M_ZIP = Path("data/ml-32m.zip")


def test_latest_small_dir():
    ds = load_movielens(ML_LATEST_DIR)
    assert ds.item_count >= 100
    assert ds.user_count >= 100
    assert ds.user_count < 1000
    assert ds.interaction_count >= 100_000

    assert ds.name == "ml-latest-small"

    titles = ds.entities("item").attribute("title")
    title_s = titles.pandas()
    assert title_s.loc[1] == "Toy Story (1995)"

    genres = ds.entities("item").attribute("genres").pandas(missing="omit")
    # Cry, The Beloved Country is drama
    assert genres.loc[40] == ["Drama"]
    assert not np.any(genres.isnull())

    # Scorpio Rising has no genres
    assert 83829 in ds.entities("item").attribute("genres").id_index()
    assert 83829 not in genres.index

    genre_nulls = ds.entities("item").attribute("genres").pandas(missing="null")
    assert len(genre_nulls) > len(genres)
    assert len(genre_nulls) == ds.item_count
    assert not np.all(genre_nulls.isnull())
    print(genre_nulls.loc[82000:83000])
    assert genre_nulls.loc[83829] == None  # noqa: E711

    assert "tag_counts" in ds.entities("item").attributes
    tags = ds.entities("item").attribute("tag_counts")
    assert tags.is_sparse
    assert tags.scipy().data.sum() >= 1200

    ratings = ds.interactions()
    assert ratings.attribute_names == ["rating", "timestamp"]
    assert set(ds.schema.relationships["rating"].attributes.keys()) == {"rating", "timestamp"}


def test_ml_modern(tmpdir: Path):
    tgt = tmpdir / "ml-football"
    copytree(ml_test_dir, tgt)

    ds = load_movielens(tgt)
    assert ds.item_count >= 100
    assert ds.user_count >= 100
    assert ds.user_count < 1000
    assert ds.interaction_count >= 100_000

    assert ds.name == "ml-modern"


@mark.skipif(not ML_100K_ZIP.exists(), reason="ml-100k does not exist")
@mark.realdata
def test_100k_zip():
    ds = load_movielens(ML_100K_ZIP)
    assert ds.name == "ml-100k"
    assert ds.item_count >= 100
    assert ds.user_count >= 100
    assert ds.user_count < 1000
    assert ds.interaction_count >= 100_000

    titles = ds.entities("item").attribute("title")
    title_s = titles.pandas()
    assert title_s.loc[1] == "Toy Story (1995)"

    genders = ds.entities("user").attribute("gender").pandas()
    # only binary gender is recorded
    assert set(genders) == {"M", "F"}


@mark.skipif(not ML_100K_ZIP.exists(), reason="ml-100k does not exist")
@mark.realdata
def test_100k_df():
    ds = load_movielens_df(ML_100K_ZIP)
    assert ds["item_id"].nunique() >= 100
    assert ds["user_id"].nunique() >= 100
    assert ds["user_id"].nunique() < 1000
    assert len(ds) >= 100_000


@mark.skipif(not ML_1M_ZIP.exists(), reason="ml-1m does not exist")
@mark.realdata
def test_1m_zip():
    ds = load_movielens(ML_1M_ZIP)
    assert ds.name == "ml-1m"
    assert ds.item_count >= 500
    assert ds.user_count >= 500
    assert ds.interaction_count >= 1_000_000

    titles = ds.entities("item").attribute("title")
    title_s = titles.pandas()
    assert title_s.loc[1] == "Toy Story (1995)"

    assert "genres" in ds.entities("item").attributes
    genres = ds.entities("item").attribute("genres").pandas()
    # Cry, The Beloved Country is drama
    assert genres.loc[40] == ["Drama"]

    genders = ds.entities("user").attribute("gender").pandas()
    # only binary gender is recorded
    assert set(genders) == {"M", "F"}


@mark.skipif(not ML_1M_ZIP.exists(), reason="ml-1m does not exist")
@mark.realdata
def test_1m_df():
    ds = load_movielens_df(ML_1M_ZIP)
    assert ds["item_id"].nunique() >= 500
    assert ds["user_id"].nunique() >= 500
    assert len(ds) >= 1_000_000


@mark.skipif(not ML_10M_ZIP.exists(), reason="ml-10m does not exist")
@mark.slow
@mark.realdata
def test_10m_zip():
    ds = load_movielens(ML_10M_ZIP)
    assert ds.name == "ml-10m"
    assert ds.item_count >= 10_000
    assert ds.user_count >= 69_000
    assert ds.interaction_count >= 10_000_000

    titles = ds.entities("item").attribute("title")
    title_s = titles.pandas()
    assert title_s.loc[1] == "Toy Story (1995)"

    genres = ds.entities("item").attribute("genres").pandas()
    # Cry, The Beloved Country is drama
    assert genres.loc[40] == ["Drama"]

    assert "tag_counts" in ds.entities("item").attributes
    tags = ds.entities("item").attribute("tag_counts")
    assert tags.is_sparse
    assert tags.scipy().data.sum() >= 95_000


@mark.skipif(not ML_10M_ZIP.exists(), reason="ml-10m does not exist")
@mark.slow
@mark.realdata
def test_10m_df():
    ds = load_movielens_df(ML_10M_ZIP)
    assert ds["item_id"].nunique() >= 10_000
    assert ds["user_id"].nunique() >= 69_000
    assert len(ds) >= 10_000_000


@mark.skipif(not ML_20M_ZIP.exists(), reason="ml-20m does not exist")
@mark.slow
@mark.realdata
def test_20m_zip():
    ds = load_movielens(ML_20M_ZIP)
    assert ds.name == "ml-20m"
    assert ds.item_count >= 25_000
    assert ds.user_count >= 130_000
    assert ds.interaction_count >= 20_000_000

    titles = ds.entities("item").attribute("title")
    title_s = titles.pandas()
    assert title_s.loc[1] == "Toy Story (1995)"

    genres = ds.entities("item").attribute("genres").pandas()
    # Cry, The Beloved Country is drama
    assert genres.loc[40] == ["Drama"]

    assert "tag_counts" in ds.entities("item").attributes
    tags = ds.entities("item").attribute("tag_counts")
    assert tags.is_sparse
    assert tags.scipy().data.sum() >= 100_000

    assert "tag_genome" in ds.entities("item").attributes
    ga = ds.entities("item").attribute("tag_genome")
    assert ga.is_vector
    assert ga.dim_names is not None
    assert ga.dim_names[0] == "007"
    gdf = ga.pandas(missing="omit")
    assert len(gdf) >= 10_000
    assert len(gdf) < 20_000
    assert gdf.columns[0] == "007"
    assert gdf.columns[-1] == "zombies"


@mark.skipif(not ML_20M_ZIP.exists(), reason="ml-20m does not exist")
@mark.slow
@mark.realdata
def test_ml20m_save(tmpdir: Path):
    ds = load_movielens(ML_20M_ZIP)

    path = tmpdir / "ml-20m"
    ds.save(path)

    assert path.exists()
    assert (path / "schema.json").exists()

    ds2 = Dataset.load(path)
    assert ds2.user_count == ds.user_count
    assert ds2.item_count == ds.item_count
    assert ds2.interaction_count == ds.interaction_count

    assert "title" in ds.entities("item").attributes
    assert ds.entities("item").attribute("title").is_scalar
    assert "genres" in ds.entities("item").attributes
    assert ds.entities("item").attribute("genres").is_list
    assert "tag_counts" in ds.entities("item").attributes
    assert ds.entities("item").attribute("tag_counts").is_sparse
    # assert "tag_genome" in ds.entities("item").attributes
    # assert ds.entities("item").attribute("tag_genome").is_vector


@mark.skipif(not ML_20M_ZIP.exists(), reason="ml-20m does not exist")
@mark.slow
@mark.realdata
def test_20m_df():
    ds = load_movielens_df(ML_20M_ZIP)
    assert ds["item_id"].nunique() >= 25_000
    assert ds["user_id"].nunique() >= 130_000
    assert len(ds) >= 20_000_000


@mark.skipif(not ML_25M_ZIP.exists(), reason="ml-25m does not exist")
@mark.slow
@mark.realdata
def test_25m_zip():
    ds = load_movielens(ML_25M_ZIP)
    assert ds.name == "ml-25m"
    assert ds.item_count >= 50_000
    assert ds.user_count >= 160_000
    assert ds.interaction_count >= 25_000_000

    titles = ds.entities("item").attribute("title")
    title_s = titles.pandas()
    assert title_s.loc[1] == "Toy Story (1995)"

    genres = ds.entities("item").attribute("genres").pandas()
    # Cry, The Beloved Country is drama
    assert genres.loc[40] == ["Drama"]

    assert "tag_counts" in ds.entities("item").attributes
    tags = ds.entities("item").attribute("tag_counts")
    assert tags.is_sparse
    assert tags.scipy().data.sum() >= 200_000

    assert "tag_genome" in ds.entities("item").attributes
    ga = ds.entities("item").attribute("tag_genome")
    assert ga.is_vector
    assert ga.dim_names is not None
    assert ga.dim_names[0] == "007"
    gdf = ga.pandas(missing="omit")
    assert len(gdf) >= 10_000
    assert len(gdf) < 20_000
    assert gdf.columns[0] == "007"
    assert gdf.columns[-1] == "zombies"


@mark.skipif(not ML_25M_ZIP.exists(), reason="ml-25m does not exist")
@mark.slow
@mark.realdata
def test_25m_df():
    ds = load_movielens_df(ML_25M_ZIP)
    assert ds["item_id"].nunique() >= 50_000
    assert ds["user_id"].nunique() >= 160_000
    assert len(ds) >= 25_000_000


@mark.skipif(not ML_32M_ZIP.exists(), reason="ml-32m does not exist")
@mark.slow
@mark.realdata
def test_32m_zip():
    ds = load_movielens(ML_32M_ZIP)
    assert ds.name == "ml-32m"
    assert ds.item_count >= 50_000
    assert ds.user_count >= 200_000
    assert ds.interaction_count >= 32_000_000

    titles = ds.entities("item").attribute("title")
    title_s = titles.pandas()
    assert title_s.loc[1] == "Toy Story (1995)"

    genres = ds.entities("item").attribute("genres").pandas()
    # Cry, The Beloved Country is drama
    assert genres.loc[40] == ["Drama"]

    assert "tag_counts" in ds.entities("item").attributes
    tags = ds.entities("item").attribute("tag_counts")
    assert tags.is_sparse
    assert tags.scipy().data.sum() >= 200_000

    # assert "tag_genome" in ds.entities("item").attributes
    # ga = ds.entities("item").attribute("tag_genome")
    # assert ga.is_vector
    # assert ga.names is not None
    # assert ga.names[0] == "007"
    # gdf = ga.pandas(missing="omit")
    # assert len(gdf) >= 10_000
    # assert len(gdf) < 20_000
    # assert gdf.columns[0] == "007"
    # assert gdf.columns[-1] == "zombies"


@mark.skipif(not ML_32M_ZIP.exists(), reason="ml-32m does not exist")
@mark.slow
@mark.realdata
def test_32m_zip_df():
    ds = load_movielens_df(ML_32M_ZIP)
    assert ds["item_id"].nunique() >= 50_000
    assert ds["user_id"].nunique() >= 200_000
    assert len(ds) >= 32_000_000


def test_bad_dir():
    with raises(RuntimeError, match="invalid.*directory"):
        load_movielens("tests")



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import numpy as np

from lenskit.data import ItemList
from lenskit.metrics import GeometricRankWeight, RankWeight


def rank_biased_overlap(
    reference: ItemList,
    reranked: ItemList,
    weight: RankWeight | None = None,
    k: int = 10,
) -> float:
    """
    Computes the Rank Biased Overlap (RBO) between two item rankings.
    RBO is defined by :cite:t:`webberSimilarityMeasureIndefinite2010`.

    Args:
        reference:
            The first item list to compare.
        reranked:
            The second item list to compare.
        weight:
            The rank weighting to use. If None, defaults to GeometricRankWeight(0.85).
        k:
            The depth to which to compute the overlap (default 10).

    Returns:
        The RBO score between 0 and 1.

    Stability:
        Experimental
    """
    if weight is None:
        weight = GeometricRankWeight(0.85)
    weights = weight.weight(np.arange(1, k + 1))

    sum = 0
    total_weights = 0

    for d, w in enumerate(weights, start=1):
        ref_ids = reference[:d].ids()
        rerank_ids = reranked[:d].ids()
        overlap = len(np.intersect1d(ref_ids, rerank_ids, assume_unique=True))
        agreement = overlap / d

        sum += agreement * w
        total_weights += w

    rbo = sum / total_weights
    return rbo



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Jupyter notebook progress support.
"""

from __future__ import annotations

from time import perf_counter

import ipywidgets as widgets
from humanize import metric
from IPython.display import display

from .._limit import RateLimit
from ._base import Progress
from ._formats import field_format

__all__ = ["JupyterProgress"]


class JupyterProgress(Progress):  # pragma: nocover
    """
    Progress logging to Jupyter notebook widgets.
    """

    widget: widgets.IntProgress
    text: widgets.Label
    box: widgets.HBox
    current: int
    _limit: RateLimit
    _field_format: str | None = None

    def __init__(
        self,
        label: str | None,
        total: int | None,
        fields: dict[str, str | None],
    ):
        super().__init__()
        self._limit = RateLimit()
        self.current = 0
        self.total = total
        if total:
            self.widget = widgets.IntProgress(value=0, min=0, max=total, step=1)
        else:
            self.widget = widgets.IntProgress(value=1, min=0, max=1, step=1)
            self.widget.bar_style = "info"

        pieces = []
        if label:
            pieces.append(widgets.Label(value=label))
        pieces.append(self.widget)

        self.text = widgets.Label()
        if total:
            self.text.value = "0 / {}".format(metric(total))
        pieces.append(self.text)

        self.box = widgets.HBox(pieces)
        display(self.box)

        if fields:
            self._field_format = ", ".join(
                [f"{name}: {field_format(name, fs)}" for (name, fs) in fields.items()]
            )

    def update(
        self,
        advance: int = 1,
        completed: int | None = None,
        total: int | None = None,
        **kwargs: float | int | str,
    ):
        """
        Update the progress bar.
        """
        if total is not None:
            self.total = total
        if completed is not None:
            self.current = completed
        else:
            self.current += advance
        now = perf_counter()
        if self._limit.want_update(now) or (self.total and self.current >= self.total):
            self.widget.value = self.current
            if self.total:
                if self.total >= 1000:
                    txt = "{} / {}".format(metric(self.current), metric(self.total))
                else:
                    txt = "{} / {}".format(self.current, self.total)
            else:
                txt = "{} / ?".format(metric(self.current))
            if self._field_format:
                txt += " ({})".format(self._field_format.format(**kwargs))
            self.text.value = txt

            self._limit.mark_update(now)
        # if self._field_format:
        #     self.tqdm.set_postfix_str(self._field_format.format(kwargs))

    def finish(self):
        """
        Finish and clean up this progress bar.  If the progresss bar is used as
        a context manager, this is automatically called on context exit.
        """
        self.box.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.finish()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle

import numpy as np

from pytest import approx, importorskip, mark

from lenskit.data import ItemList, from_interactions_df
from lenskit.metrics import quick_measure_model
from lenskit.testing import BasicComponentTests, ScorerTests

imp = importorskip("lenskit.implicit")
_log = logging.getLogger(__name__)


class TestImplicitALS(BasicComponentTests, ScorerTests):
    component = imp.ALS
    expected_ndcg = 0.05


class TestImplicitBPR(BasicComponentTests, ScorerTests):
    component = imp.BPR
    expected_ndcg = 0.05


@mark.slow
def test_implicit_als_train_rec(ml_ds):
    algo = imp.ALS(factors=25)
    assert algo.config.factors == 25

    ret = algo.train(ml_ds)
    assert ret is algo

    preds = algo(100, items=ItemList([1, 377]))
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert np.all(preds.notna())

    _log.info("serializing implicit model")
    mod = pickle.dumps(algo)
    _log.info("serialized to %d bytes")
    a2 = pickle.loads(mod)

    p2 = a2(100, items=ItemList([1, 377])).scores("pandas", index="ids")
    assert np.all(p2 == preds)


@mark.slow
def test_implicit_bpr_train_rec(ml_ds):
    algo = imp.BPR(factors=25, use_gpu=False)
    assert algo.config.factors == 25

    algo.train(ml_ds)

    preds = algo(100, ItemList([20, 30, 23148010]))
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert all(preds.index == [20, 30, 23148010])
    assert all(preds.isna() == [False, False, True])

    _log.info("serializing implicit model")
    mod = pickle.dumps(algo)
    _log.info("serialized to %d bytes")
    a2 = pickle.loads(mod)

    p2 = a2(100, items=ItemList([20, 30, 23148010])).scores("pandas", index="ids")
    assert p2.values == approx(preds.values, nan_ok=True)


def test_implicit_pickle_untrained(tmp_path):
    mf = tmp_path / "bpr.dat"
    algo = imp.BPR(factors=25, use_gpu=False)

    with mf.open("wb") as f:
        pickle.dump(algo, f)

    with mf.open("rb") as f:
        a2 = pickle.load(f)

    assert a2 is not algo
    assert a2.config.factors == 25



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from typing import Any

from lenskit.data.dataset import Dataset
from lenskit.data.vocab import Vocabulary
from lenskit.pipeline import PipelineBuilder
from lenskit.training import Trainable, TrainingOptions


def test_train(ml_ds: Dataset):
    pipe = PipelineBuilder()
    item = pipe.create_input("item", int)

    tc: Trainable = TestComponent()
    pipe.add_component("test", tc, item=item)
    pipe.default_component("test")

    pipe = pipe.build()
    pipe.train(ml_ds)

    # return true for an item that exists
    assert pipe.run(item=500)
    # return false for an item that does not
    assert not pipe.run(item=-100)


class TestComponent:
    items: Vocabulary

    def __call__(self, *, item: int) -> bool:
        return self.items.number(item, "none") is not None

    def train(self, data: Dataset, options: TrainingOptions):
        # we just memorize the items
        self.items = data.items

    def get_params(self) -> dict[str, object]:
        return {"items": self.items}

    def load_params(self, params: dict[str, Any]) -> None:
        self.items = params["items"]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle
import warnings
from pathlib import Path

import numpy as np
import pandas as pd
import pyarrow as pa

from pytest import mark, raises, warns

from lenskit.data import ItemList
from lenskit.data.collection import ItemListCollection, MutableItemListCollection, UserIDKey
from lenskit.data.collection._keys import create_key, project_key
from lenskit.data.dataset import Dataset
from lenskit.diagnostics import DataWarning
from lenskit.testing import DemoRecs, demo_recs

_log = logging.getLogger(__name__)


def test_generic_key():
    usk = create_key(("user_id", "seq_no"), "alphabet", 42)

    assert isinstance(usk, tuple)
    assert usk == ("alphabet", 42)
    assert usk.user_id == "alphabet"  # type: ignore
    assert usk.seq_no == 42  # type: ignore


def test_pickle_generic():
    usk = create_key(("user_id", "seq_no"), "alphabet", 42)

    bs = pickle.dumps(usk)
    usk2 = pickle.loads(bs)

    assert isinstance(usk2, tuple)
    assert usk2 == usk


def test_project_key():
    usk = create_key(("user_id", "seq_no"), "alphabet", 42)

    uk = project_key(usk, UserIDKey)
    assert isinstance(uk, UserIDKey)
    assert uk.user_id == "alphabet"
    assert uk == ("alphabet",)


def test_project_missing_fails():
    usk = create_key(("seq_no",), 42)

    with raises(TypeError, match="missing field"):
        project_key(usk, UserIDKey)


def test_collection_empty():
    empty = ItemListCollection.empty(UserIDKey)
    assert len(empty) == 0
    with raises(StopIteration):
        next(iter(empty))


def test_collection_base_ctor():
    empty = ItemListCollection(UserIDKey)  # type: ignore
    assert len(empty) == 0
    with raises(StopIteration):
        next(iter(empty))


def test_collection_mutable_ctor():
    empty = MutableItemListCollection(UserIDKey)  # type: ignore
    assert len(empty) == 0
    with raises(StopIteration):
        next(iter(empty))


def test_collection_add_item():
    ilc = ItemListCollection.empty(UserIDKey)

    ilc.add(ItemList(["a"]), 72)
    assert len(ilc) == 1
    key, il = next(iter(ilc))
    assert key.user_id == 72
    assert np.all(il.ids() == ["a"])

    il2 = ilc.lookup(72)
    assert il2 is il


def test_collection_from_dict_nt_key():
    ilc = ItemListCollection.from_dict({UserIDKey(72): ItemList(["a"])}, key=UserIDKey)

    assert len(ilc) == 1
    k, il = ilc[0]
    assert k.user_id == 72
    assert len(il) == 1
    assert np.all(il.ids() == ["a"])


def test_collection_from_dict_tuple_key():
    ilc = ItemListCollection.from_dict({(72,): ItemList(["a"])}, key=UserIDKey)

    assert len(ilc) == 1
    k, il = ilc[0]
    assert k.user_id == 72
    assert len(il) == 1
    assert np.all(il.ids() == ["a"])


def test_collection_from_dict_singleton_key():
    ilc = ItemListCollection.from_dict({72: ItemList(["a"])}, key=UserIDKey)

    assert len(ilc) == 1
    k, il = ilc[0]
    assert k.user_id == 72
    assert len(il) == 1
    assert np.all(il.ids() == ["a"])


def test_collection_from_dict_singleton_field():
    ilc = ItemListCollection.from_dict({72: ItemList(["a"])}, key="user_id")

    assert len(ilc) == 1
    k, il = ilc[0]
    assert k.user_id == 72  # type: ignore
    assert len(il) == 1
    assert np.all(il.ids() == ["a"])


def test_lookup_projected():
    ilc = ItemListCollection.from_dict({72: ItemList(["a"])}, key="user_id")

    usk = create_key(("user_id", "seq"), 72, 100)
    il = ilc.lookup_projected(usk)
    assert il is not None
    assert len(il) == 1
    assert np.all(il.ids() == ["a"])


def test_add_from():
    ilc = ItemListCollection.empty(["model", "user_id"])

    ilc1 = ItemListCollection.from_dict({72: ItemList(["a", "b"]), 48: ItemList()}, key="user_id")
    ilc.add_from(ilc1, model="foo")

    assert len(ilc) == 2
    il = ilc.lookup(("foo", 72))
    assert il is not None
    assert il.ids().tolist() == ["a", "b"]
    il = ilc.lookup(("foo", 48))
    assert il is not None
    assert len(il) == 0


def test_from_df(rng, ml_ratings: pd.DataFrame):
    ml_ratings = ml_ratings.rename(columns={"user": "user_id", "item": "item_id"})
    ilc = ItemListCollection.from_df(ml_ratings, UserIDKey)
    assert len(ilc) == ml_ratings["user_id"].nunique()
    assert set(k.user_id for k in ilc.keys()) == set(ml_ratings["user_id"])

    for uid in rng.choice(ml_ratings["user_id"].unique(), 25):
        items = ilc.lookup(user_id=uid)
        udf = ml_ratings[ml_ratings["user_id"] == uid]
        assert len(items) == len(udf)
        assert np.all(np.unique(items.ids()) == np.unique(udf["item_id"]))

    tot = sum(len(il) for il in ilc.lists())
    assert tot == len(ml_ratings)


def test_from_df_auto(rng, ml_ratings: pd.DataFrame):
    ml_ratings = ml_ratings.rename(columns={"user": "user_id", "item": "item_id"})
    with warns(DataWarning, match="inferring"):
        ilc = ItemListCollection.from_df(ml_ratings)

    assert len(ilc) == ml_ratings["user_id"].nunique()
    assert set(k.user_id for k in ilc.keys()) == set(ml_ratings["user_id"])

    for uid in rng.choice(ml_ratings["user_id"].unique(), 25):
        items = ilc.lookup(user_id=uid)
        udf = ml_ratings[ml_ratings["user_id"] == uid]
        assert len(items) == len(udf)
        assert np.all(np.unique(items.ids()) == np.unique(udf["item_id"]))

    tot = sum(len(il) for il in ilc.lists())
    assert tot == len(ml_ratings)


def test_to_df():
    ilc = ItemListCollection.from_dict(
        {72: ItemList(["a"], scores=[1]), 82: ItemList(["a", "b", "c"], scores=[3, 4, 10])},
        key="user_id",
    )
    df = ilc.to_df()
    print(df)
    assert len(df) == 4
    assert df["user_id"].tolist() == [72, 82, 82, 82]
    assert list(df.columns) == ["user_id", "item_id", "score"]


def test_to_df_no_nums():
    ilc = ItemListCollection.from_dict(
        {
            72: ItemList(["a"], item_nums=[0], scores=[1]),
            82: ItemList(["a", "b", "c"], item_nums=np.arange(3), scores=[3, 4, 10]),
        },
        key="user_id",
    )
    df = ilc.to_df()
    print(df)
    assert len(df) == 4
    assert df["user_id"].tolist() == [72, 82, 82, 82]
    assert list(df.columns) == ["user_id", "item_id", "score"]


def test_to_df_warn_empty():
    ilc = ItemListCollection.from_dict(
        {
            72: ItemList(["a"], scores=[1]),
            40: ItemList(),
            82: ItemList(["a", "b", "c"], scores=[3, 4, 10]),
        },
        key="user_id",
    )
    with warns(DataWarning, match="dropped"):
        df = ilc.to_df()
    print(df)
    assert len(df) == 4
    assert df["user_id"].tolist() == [72, 82, 82, 82]


def test_to_arrow():
    ilc = ItemListCollection.from_dict(
        {72: ItemList(["a"], scores=[1]), 82: ItemList(["a", "b", "c"], scores=[3, 4, 10])},
        key="user_id",
    )
    tbl = ilc.to_arrow()
    print(tbl)
    assert tbl.num_rows == 2
    assert np.all(tbl.column("user_id").to_numpy() == [72, 82])
    il_field = tbl.field("items")
    assert pa.types.is_list(il_field.type)
    assert isinstance(il_field.type, pa.ListType)
    il_type = il_field.type.value_type
    assert pa.types.is_struct(il_type)
    assert isinstance(il_type, pa.StructType)
    names = [il_type.field(i).name for i in range(il_type.num_fields)]
    assert names == ["item_id", "score"]


def test_to_arrow_flat():
    ilc = ItemListCollection.from_dict(
        {72: ItemList(["a"], scores=[1]), 82: ItemList(["a", "b", "c"], scores=[3, 4, 10])},
        key="user_id",
    )
    tbl = ilc.to_arrow(layout="flat")
    print(tbl)
    assert tbl.num_columns == 3
    assert tbl.num_rows == 4
    assert np.all(tbl.column("user_id").to_numpy() == [72, 82, 82, 82])
    assert np.all(tbl.column("item_id").to_numpy() == ["a", "a", "b", "c"])
    assert np.all(tbl.column("score").to_numpy() == [1, 3, 4, 10])


@mark.parametrize("layout", ["native", "flat"])
def test_save_parquet(ml_ds: Dataset, tmpdir: Path, layout):
    ilc = ItemListCollection.empty(["user_id"])
    for user in ml_ds.users.ids():
        ilc.add(ml_ds.user_row(user), user_id=user)

    _log.info("initial list:\n%s", ilc.to_df())

    f = tmpdir / "items.parquet"
    ilc.save_parquet(f, layout=layout)

    assert f.exists()

    ilc2 = ItemListCollection.load_parquet(f, layout=layout)
    _log.info("loaded list:\n%s", ilc2.to_df())
    assert len(ilc2) == len(ilc)
    assert set(ilc2.keys()) == set(ilc.keys())

    il1 = next(ilc.lists())
    _log.info("first item (initial):\n%s", il1.to_df())
    il2 = next(ilc2.lists())
    _log.info("first item (loaded):\n%s", il2.to_df())
    assert len(il1) == len(il2)

    assert sum(len(l1) for l1 in ilc2.lists()) == sum(len(l2) for l2 in ilc.lists())


@mark.filterwarnings("ignore:.*dropped.*:lenskit.diagnostics.DataWarning")
def test_save_parquet_with_empty(ml_ds: Dataset, tmpdir: Path):
    ilc = ItemListCollection.empty(["user_id"])
    ilc.add(ItemList(), user_id=-1)
    for user in ml_ds.users.ids():
        if user % 348:
            ilc.add(ml_ds.user_row(user), user_id=user)
        else:
            ilc.add(ItemList(), user_id=user)

    _log.info("initial list:\n%s", ilc.to_df())

    f = tmpdir / "items.parquet"
    ilc.save_parquet(f)

    assert f.exists()

    ilc2 = ItemListCollection.load_parquet(f)
    _log.info("loaded list:\n%s", ilc2.to_df())
    assert len(ilc2) == len(ilc)
    assert set(ilc2.keys()) == set(ilc.keys())

    il1 = next(ilc.lists())
    _log.info("first item (initial):\n%s", il1.to_df())
    il2 = next(ilc2.lists())
    _log.info("first item (loaded):\n%s", il2.to_df())
    assert len(il1) == len(il2)

    assert sum(len(l1) for l1 in ilc2.lists()) == sum(len(l2) for l2 in ilc.lists())


def test_save_parquet_with_mkdir(tmpdir: Path):
    ilc = ItemListCollection.empty(["user_id"])

    f = tmpdir / "subdir" / "items.parquet"
    ilc.save_parquet(f, mkdir=True)
    assert (tmpdir / "subdir").exists()

    f_no_mkdir = tmpdir / "no_mkdir" / "items.parquet"
    ilc.save_parquet(f_no_mkdir, mkdir=False)
    assert not (tmpdir / "no_mkdir").exists()


def test_write_recs_parquet(demo_recs, tmpdir: Path):
    split, recs = demo_recs

    test_f = tmpdir / "test.parquet"
    rec_f = tmpdir / "recs.parquet"

    split.test.save_parquet(test_f)
    recs.save_parquet(rec_f)

    t2 = ItemListCollection.load_parquet(test_f)
    assert list(t2.keys()) == list(split.test.keys())

    r2 = ItemListCollection.load_parquet(rec_f)
    assert list(r2.keys()) == list(recs.keys())
    assert all(il.ordered for il in r2.lists())


def test_recs_df_expected_column(demo_recs: DemoRecs):
    rec_df = demo_recs.recommendations.to_df()
    print(rec_df)
    print(demo_recs.recommendations[0])
    assert list(rec_df.columns) == ["user_id", "item_id", "score", "rank"]


def test_to_dataset(demo_recs: DemoRecs):
    test = demo_recs.split.test
    test_ds = test.to_dataset()
    assert test_ds.user_count == len(test)
    assert set(test_ds.users.ids()) == set(k.user_id for k in test.keys())

    test_df = test.to_df()
    assert test_ds.interaction_count == len(test_df)

    src_user_counts = test_df.groupby("user_id")["item_id"].count()
    src_item_counts = test_df.groupby("item_id")["user_id"].count()

    ds_user_counts = test_ds.user_stats()["count"]
    ds_item_counts = test_ds.item_stats()["count"]

    src_user_counts, ds_user_counts = src_user_counts.align(ds_user_counts, fill_value=0)
    src_item_counts, ds_item_counts = src_item_counts.align(ds_item_counts, fill_value=0)

    assert np.all(ds_user_counts == src_user_counts)
    assert np.all(ds_item_counts == src_item_counts)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Hierarchical Poisson factorization from ``hpfrec``.
"""

import logging

import hpfrec
import numpy as np
from pydantic import AliasChoices, BaseModel, Field, JsonValue
from typing_extensions import override

from lenskit.data import Dataset, ItemList, QueryInput, RecQuery, Vocabulary
from lenskit.pipeline import Component
from lenskit.training import Trainable, TrainingOptions

_logger = logging.getLogger(__name__)


class HPFConfig(BaseModel, extra="allow"):
    """
    Configuration class for :class:`HPFScorer`.

    Only a couple of configuration options are directly defined here.  Other
    options are taken from the :class:`hpfrec.HPF` constructor's keyword
    arguments.
    """

    __pydantic_extra__: dict[str, JsonValue]
    count_attribute: str | None = "count"
    """
    The attribute from which to get user-item interaction counts.  If ``None``,
    uses indicator variables as a count of 1.
    """

    embedding_size: int = Field(
        default=50, validation_alias=AliasChoices("embedding_size", "features")
    )
    """
    The dimension of user and item embeddings (number of latent features to
    learn).
    """


class HPFScorer(Component[ItemList], Trainable):
    """
    Hierarchical Poisson factorization, provided by `hpfrec`_

    .. _hpfrec: https://hpfrec.readthedocs.io/en/latest/

    .. todo::
        Right now, this uses the 'rating' as a count. Actually use counts (:issue:`656`).

    Stability:
        Experimental

    Args:
        features:
            the number of features
        kwargs:
            additional arguments to pass to :class:`hpfrec.HPF`.
    """

    config: HPFConfig

    users_: Vocabulary
    user_features_: np.ndarray[tuple[int, int], np.dtype[np.float64]]
    items_: Vocabulary
    item_features_: np.ndarray[tuple[int, int], np.dtype[np.float64]]

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        if hasattr(self, "item_features_") and not options.retrain:
            return

        interacts = data.interactions().matrix()
        matrix = interacts.scipy(self.config.count_attribute, layout="coo")

        hpf = hpfrec.HPF(
            self.config.embedding_size, reindex=False, **self.config.__pydantic_extra__
        )  # type: ignore

        _logger.info("fitting HPF model with %d features", self.config.embedding_size)
        hpf.fit(matrix)

        self.users_ = data.users
        self.items_ = data.items
        self.user_features_ = hpf.Theta  # type: ignore
        self.item_features_ = hpf.Beta  # type: ignore

        return self

    @override
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        query = RecQuery.create(query)

        user_id = query.user_id
        user_num = None
        if user_id is not None:
            user_num = self.users_.number(user_id, missing=None)
        if user_num is None:
            _logger.debug("unknown user %s", query.user_id)
            return ItemList(items, scores=np.nan)

        u_feat = self.user_features_[user_num, :]

        item_nums = items.numbers(vocabulary=self.items_, missing="negative")
        item_mask = item_nums >= 0
        i_feats = self.item_features_[item_nums[item_mask], :]

        scores = np.full((len(items),), np.nan, dtype=np.float64)
        scores[item_mask] = i_feats @ u_feat

        return ItemList(items, scores=scores)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd

import pytest

from lenskit.basic import PopScorer
from lenskit.data import DatasetBuilder, ItemList, from_interactions_df
from lenskit.logging import get_logger
from lenskit.pipeline import RecPipelineBuilder
from lenskit.reranking.fair import FAIRReranker, FAIRRerankerConfig
from lenskit.training import TrainingOptions

_log = get_logger(__name__)


def make_dataset(item_ids, protected_flags, protected_attribute="protected"):
    """Helper to build a minimal Dataset with protected attributes."""

    # dummy interactions dataset
    interactions = pd.DataFrame(
        {
            "user_id": range(len(item_ids)),
            "item_id": item_ids,
        }
    )

    protected_df = pd.DataFrame(
        {
            "item_id": item_ids,
            protected_attribute: protected_flags,
        }
    )

    ds = from_interactions_df(interactions)
    builder = DatasetBuilder(ds)

    builder.add_scalar_attribute("item", protected_attribute, protected_df)
    ds = builder.build()
    return ds


def test_compute_m_list():
    n, p, alpha = 10, 0.5, 0.1
    reranker = FAIRReranker(n=n, p=p, alpha=alpha)
    m_list = reranker._compute_m_list(n, p, alpha)
    # check Table 2 of Zehlike et al. (2017)
    assert m_list[3] == 1
    assert m_list[6] == 2
    assert m_list[8] == 3
    assert len(m_list) == n


def test_compute_blocks():
    reranker = FAIRReranker(n=9, p=0.5, alpha=0.1)
    m_list = np.array([0, 0, 0, 1, 1, 1, 2, 2, 3])
    blocks = reranker._compute_blocks(m_list)

    assert list(blocks) == [4, 3, 2]


def test_rerank_small_list():
    items = np.arange(6, 15)
    protected = [False, False, False, True, False, True, True, False, True]
    data = make_dataset(items, protected, protected_attribute="flag")

    reranker = FAIRReranker(n=9, p=0.5, alpha=0.1, protected_attribute="flag")
    reranker.train(data)

    sample_vanilla_list = ItemList(item_ids=items, ordered=True)
    fair_reranked = reranker(sample_vanilla_list)

    # Check that protected counts meet m_list thresholds
    ids = fair_reranked.ids()
    prot_flags = data.entities("item").attribute("flag").pandas().reindex(ids)

    counts = prot_flags.cumsum()
    assert np.all(counts >= reranker.m_list_[: len(ids)])


@pytest.mark.parametrize(
    "protected",
    [
        [False] * 6,  # all unprotected
        [True] * 6,  # all protected
    ],
)
def test_all_unprotected_items(protected):
    items = np.arange(6, 12)
    data = make_dataset(items, protected)

    reranker = FAIRReranker(n=6, p=0.5, alpha=0.1)
    reranker.train(data)

    sample_vanilla_list = ItemList(item_ids=items, ordered=True)
    fair_reranked = reranker(sample_vanilla_list)

    # Nothing to rerank, should return same order
    assert np.array_equal(fair_reranked.ids(), items)


@pytest.mark.parametrize(
    "n,p,alpha", [(20, 0.5, 0.2), (50, 0.5, 0.1), (100, 0.1, 0.3), (150, 0.9, 0.1)]
)
def test_randomized_reranking(n, p, alpha):
    rng = np.random.default_rng(42)
    items = rng.permutation(n)
    protected = rng.random(n) < p
    data = make_dataset(items, protected)

    reranker = FAIRReranker(n=n, p=p, alpha=alpha)
    reranker.train(data)

    sample_vanilla_list = ItemList(item_ids=items, ordered=True)
    fair_reranked = reranker(sample_vanilla_list)

    # verify that fairness constraints are enforced
    counts = np.cumsum([iid in items[protected] for iid in fair_reranked.ids()])
    assert np.all(counts >= reranker.m_list_[: len(counts)])


def test_pipeline_with_fair_reranker():
    """
    End-to-end test: PopScorer -> FAIRReranker in a RecPipelineBuilder.
    Ensures reranker can work as part of the pipeline.
    """
    items = np.arange(5, 15)
    protected = [False, False, False, False, False, True, False, False, True, True]
    ds = make_dataset(items, protected)
    runtime_n = 8
    config_n = 9

    builder = RecPipelineBuilder()
    builder.scorer(PopScorer)
    builder.reranker(FAIRReranker, FAIRRerankerConfig(n=config_n, alpha=0.2))  # attach reranker

    pipe = builder.build()
    pipe.train(ds, TrainingOptions())

    # get recommendations for a user
    recs = pipe.run("recommender", query=0, n=runtime_n)
    assert isinstance(recs, ItemList)
    ids = recs.ids()
    assert len(ids) == runtime_n

    # verify that fairness constraints are enforced
    prot_flags = ds.entities("item").attribute("protected").pandas().reindex(ids)
    counts = prot_flags.cumsum()
    reranker = pipe.node("reranker").component
    assert np.all(counts >= reranker.m_list_[: len(ids)])



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle
from pathlib import Path

import numpy as np
import pandas as pd

from pytest import approx, importorskip, mark

from lenskit import batch
from lenskit.data import Dataset, ItemList, ItemListCollection, UserIDKey, from_interactions_df
from lenskit.pipeline.common import predict_pipeline
from lenskit.testing import BasicComponentTests, ScorerTests

funk = importorskip("lenskit.funksvd")

_log = logging.getLogger(__name__)

simple_df = pd.DataFrame(
    {"item": [1, 1, 2, 3], "user": [10, 12, 10, 13], "rating": [4.0, 3.0, 5.0, 2.0]}
)
simple_ds = from_interactions_df(simple_df)


class TestFunkSVD(BasicComponentTests, ScorerTests):
    component = funk.FunkSVDScorer
    expected_rmse = (0.87, 0.97)

    def verify_models_equivalent(self, orig, copy):
        assert copy.bias.global_bias == orig.bias.global_bias
        assert np.all(copy.bias.user_biases == orig.bias.user_biases)
        assert np.all(copy.bias.item_biases == orig.bias.item_biases)
        assert np.all(copy.user_embeddings == orig.user_embeddings)
        assert np.all(copy.item_embeddings == orig.item_embeddings)
        assert np.all(copy.items.index == orig.items.index)
        assert np.all(copy.users.index == orig.users.index)


def test_fsvd_basic_build():
    algo = funk.FunkSVDScorer(features=20, epochs=20)
    assert algo.config is not None
    assert algo.config.embedding_size == 20
    algo.train(simple_ds)

    assert algo.bias is not None
    assert algo.bias.global_bias == approx(simple_df.rating.mean())
    assert algo.item_embeddings.shape == (3, 20)
    assert algo.user_embeddings.shape == (3, 20)


def test_fsvd_clamp_build():
    algo = funk.FunkSVDScorer(features=20, epochs=20, range=(1, 5))
    algo.train(simple_ds)

    assert algo.bias is not None
    assert algo.bias.global_bias == approx(simple_df.rating.mean())
    assert algo.item_embeddings.shape == (3, 20)
    assert algo.user_embeddings.shape == (3, 20)


def test_fsvd_predict_basic():
    algo = funk.FunkSVDScorer(features=20, epochs=20)
    algo.train(simple_ds)

    assert algo.bias is not None
    assert algo.bias.global_bias == approx(simple_df.rating.mean())
    assert algo.item_embeddings.shape == (3, 20)
    assert algo.user_embeddings.shape == (3, 20)

    preds = algo(query=10, items=ItemList([3]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 3
    assert preds.loc[3] >= 0
    assert preds.loc[3] <= 5


def test_fsvd_predict_clamp():
    algo = funk.FunkSVDScorer(features=20, epochs=20, range=(1, 5))
    algo.train(simple_ds)

    assert algo.bias is not None
    assert algo.bias.global_bias == approx(simple_df.rating.mean())
    assert algo.item_embeddings.shape == (3, 20)
    assert algo.user_embeddings.shape == (3, 20)

    preds = algo(query=10, items=ItemList([3]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 3
    assert preds.loc[3] >= 1
    assert preds.loc[3] <= 5


def test_fsvd_predict_bad_item():
    algo = funk.FunkSVDScorer(features=20, epochs=20)
    algo.train(simple_ds)

    assert algo.bias is not None
    assert algo.bias.global_bias == approx(simple_df.rating.mean())
    assert algo.item_embeddings.shape == (3, 20)
    assert algo.user_embeddings.shape == (3, 20)

    preds = algo(10, ItemList([4]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 4
    assert np.isnan(preds.loc[4])


def test_fsvd_predict_bad_item_clamp():
    algo = funk.FunkSVDScorer(features=20, epochs=20, range=(1, 5))
    algo.train(simple_ds)

    assert algo.bias is not None
    assert algo.bias.global_bias == approx(simple_df.rating.mean())
    assert algo.item_embeddings.shape == (3, 20)
    assert algo.user_embeddings.shape == (3, 20)

    preds = algo(10, ItemList([4]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 4
    assert np.isnan(preds.loc[4])


def test_fsvd_predict_bad_user():
    algo = funk.FunkSVDScorer(features=20, epochs=20)
    algo.train(simple_ds)

    assert algo.bias is not None
    assert algo.bias.global_bias == approx(simple_df.rating.mean())
    assert algo.item_embeddings.shape == (3, 20)
    assert algo.user_embeddings.shape == (3, 20)

    preds = algo(query=50, items=ItemList([3]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 3
    assert np.isnan(preds.loc[3])


@mark.slow
def test_fsvd_known_preds(ml_ds: Dataset):
    algo = funk.FunkSVDScorer(features=15, epochs=125, lrate=0.001)
    _log.info("training %s on ml data", algo)
    pipe = predict_pipeline(algo, fallback=False)
    pipe.train(ml_ds)

    dir = Path(__file__).parent
    pred_file = dir / "funksvd-preds.csv"
    _log.info("reading known predictions from %s", pred_file)
    known_preds = pd.read_csv(str(pred_file))
    known = ItemListCollection.from_df(known_preds, UserIDKey)

    preds = batch.predict(pipe, known, n_jobs=1)
    preds = preds.to_df().drop(columns=["prediction"], errors="ignore")

    known_preds.rename(columns={"prediction": "expected"}, inplace=True)
    merged = pd.merge(known_preds, preds)

    merged["error"] = merged.expected - merged.score
    assert not any(merged.score.isna() & merged.expected.notna())
    err = merged.error
    err = err[err.notna()]
    try:
        assert all(err.abs() < 0.01)
    except AssertionError as e:
        bad = merged[merged.error.notna() & (merged.error.abs() >= 0.01)]
        _log.error("erroneous predictions:\n%s", bad)
        raise e



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Logging, progress, and resource records.
"""

from ._console import console, stdout_console
from ._proxy import get_logger
from .config import LoggingConfig, basic_logging, notebook_logging
from .formats import friendly_duration
from .progress import Progress, item_progress, set_progress_impl
from .stopwatch import Stopwatch
from .tasks import Task
from .tracing import Tracer, get_tracer, trace

__all__ = [
    "LoggingConfig",
    "basic_logging",
    "notebook_logging",
    "Progress",
    "item_progress",
    "set_progress_impl",
    "Task",
    "get_logger",
    "get_tracer",
    "trace",
    "Tracer",
    "console",
    "stdout_console",
    "friendly_duration",
    "Stopwatch",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from ._base import Progress
from ._dispatch import item_progress, set_progress_impl
from ._handles import item_progress_handle, pbh_update

__all__ = [
    "Progress",
    "set_progress_impl",
    "item_progress",
    "item_progress_handle",
    "pbh_update",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import pickle
from datetime import datetime, timedelta

import numpy as np
import pandas as pd

from lenskit.basic import popularity
from lenskit.data import from_interactions_df
from lenskit.data.items import ItemList
from lenskit.testing import ScorerTests

ts = datetime(year=2024, month=1, day=1)
one_day_ago = ts - timedelta(days=1)
two_days_ago = ts - timedelta(days=2)
simple_df = pd.DataFrame(
    {
        "item_id": [1, 2, 2, 3],
        "user_id": [10, 12, 10, 13],
        "rating": [4.0, 3.0, 5.0, 2.0],
        "timestamp": [ts, one_day_ago, one_day_ago, one_day_ago],
    }
)
simple_ds = from_interactions_df(simple_df)


class TestTimeBoundedPop(ScorerTests):
    component = popularity.TimeBoundedPopScore
    config = popularity.TimeBoundedPopConfig(cutoff="2015-01-01")

    def verify_models_equivalent(self, orig, copy):
        assert all(orig.item_scores_ == copy.item_scores_)


def test_time_bounded_pop_score_quantile_one_day_window():
    algo = popularity.TimeBoundedPopScore(cutoff=one_day_ago)
    algo.train(simple_ds)
    assert np.all(algo.item_scores_ == [1.0, 0.0, 0.0])


def test_time_bounded_pop_score_quantile_one_day_window_call_interface():
    algo = popularity.TimeBoundedPopScore(cutoff=one_day_ago)
    algo.train(simple_ds)
    p = algo(ItemList(item_ids=[1, 2, 3]))

    assert len(p) == 3
    assert (p.scores() == [1.0, 0.0, 0.0]).all()


def test_time_bounded_pop_score_quantile_two_day_window():
    algo = popularity.TimeBoundedPopScore(cutoff=two_days_ago)
    algo.train(simple_ds)
    assert np.all(algo.item_scores_ == pd.Series([0.25, 1.0, 0.5], index=[1, 2, 3]))


def test_time_bounded_pop_score_fallbacks_to_pop_score_for_dataset_without_timestamps():
    ds = from_interactions_df(simple_df.drop(columns=["timestamp"]))

    algo = popularity.TimeBoundedPopScore(cutoff=one_day_ago)
    algo.train(ds)
    assert np.all(algo.item_scores_ == pd.Series([0.25, 1.0, 0.5], index=[1, 2, 3]))


def test_time_bounded_pop_score_rank():
    algo = popularity.TimeBoundedPopScore(cutoff=two_days_ago, score="rank")
    algo.train(simple_ds)
    assert np.all(algo.item_scores_ == pd.Series([1.5, 3.0, 1.5], index=[1, 2, 3]))


def test_time_bounded_pop_score_counts():
    algo = popularity.TimeBoundedPopScore(cutoff=two_days_ago, score="count")
    algo.train(simple_ds)
    assert np.all(algo.item_scores_ == pd.Series([1, 2, 1], index=[1, 2, 3], dtype=np.int32))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pydantic import BaseModel

from lenskit.data import ItemList
from lenskit.pipeline import Component


class LinearBlendConfig(BaseModel):
    "Configuration for :class:`LinearBlendScorer`."

    # define the parameter with a type, default value, and docstring.
    mix_weight: float = 0.5
    r"""
    Linear blending mixture weight :math:`\alpha`.
    """


class LinearBlendScorer(Component[ItemList]):
    r"""
    Score items with a linear blend of two other scores.

    Given a mixture weight :math:`\alpha` and two scores
    :math:`s_i^{\mathrm{left}}` and :math:`s_i^{\mathrm{right}}`, this
    computes :math:`s_i = \alpha s_i^{\mathrm{left}} + (1 - \alpha)
    s_i^{\mathrm{right}}`.  Missing values propagate, so only items
    scored in both inputs have scores in the output.
    """

    # define the configuration attribute, with a docstring to make sure
    # it shows up in component docs.
    config: LinearBlendConfig
    "Configuration parameters for the linear blend."

    # the __call__ method defines the component's operation
    def __call__(self, left: ItemList, right: ItemList) -> ItemList:
        """
        Blend the scores of two item lists.
        """
        ls = left.scores("pandas", index="ids")
        rs = right.scores("pandas", index="ids")
        ls, rs = ls.align(rs)
        alpha = self.config.mix_weight
        combined = ls * alpha + rs * (1 - alpha)
        return ItemList(item_ids=combined.index, scores=combined.values)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Hyperparameter searching wrapper.

.. stability:: experimental

"""

from pathlib import Path
from typing import Any

import numpy as np
import ray.tune
import ray.tune.schedulers
import ray.tune.search
from pydantic import JsonValue
from ray.tune.search.hyperopt import HyperOptSearch
from ray.tune.search.optuna import OptunaSearch

from lenskit.config import TuneSettings, lenskit_config
from lenskit.data import Dataset, ItemListCollection
from lenskit.data.collection._keys import GenericKey
from lenskit.logging import get_logger
from lenskit.parallel import get_parallel_config
from lenskit.parallel.ray import ensure_cluster
from lenskit.pipeline import PipelineBuilder, PipelineConfig
from lenskit.pipeline.nodes import ComponentConstructorNode
from lenskit.random import RNGInput, default_rng, int_seed, spawn_seed
from lenskit.splitting import TTSplit
from lenskit.training import UsesTrainer

from ._iterative import IterativeEval
from ._job import TuningJobData
from ._reporting import ProgressReport, StatusCallback
from ._simple import SimplePointEval
from ._stopper import RelativePlateauStopper
from .spec import PipelineFile, SearchSpace, TuningSpec

_log = get_logger(__name__)


class PipelineTuner:
    """
    Set up and run a hyperparameter tuning job for a pipeline.

    Stability:
        Experimental
    """

    settings: TuneSettings
    spec: TuningSpec
    out_dir: Path
    pipe_name: str | None
    random_seed: np.random.SeedSequence
    iterative: bool

    data: TTSplit[GenericKey]
    harness: Any
    tuner: ray.tune.Tuner
    """
    The Ray tuner that is used for tuning.  Not available until :meth:`setup`
    has been called.
    """
    results: ray.tune.ResultGrid
    """
    Ray tuning results. Only available after :meth:`run` has been called.
    """

    def __init__(
        self,
        spec: TuningSpec,
        out_dir: Path | None = None,
        rng: RNGInput = None,
    ):
        cfg = lenskit_config()
        self.settings = cfg.tune
        if out_dir is None:
            out_dir = Path("lenskit-tune")
        self.out_dir = out_dir

        if isinstance(spec.pipeline, PipelineFile):
            pipe_file = spec.pipeline.file
            pipe_file = spec.resolve_path(pipe_file)
            pb = PipelineBuilder.load_config(pipe_file)
        else:
            pb = PipelineBuilder.from_config(spec.pipeline, file_path=spec.file_path)

        self.spec = spec.model_copy(deep=True)
        self.spec.pipeline = pb.build_config()
        self.random_seed = spawn_seed(rng)

        self.log = _log.bind(model=pb.name)

        comp_name = self.spec.component_name
        if comp_name is None:
            self.log.error("multi-component search is not yet supported")
            raise NotImplementedError()

        comp = pb.node(comp_name)
        match comp:
            case ComponentConstructorNode(constructor=c):
                self.iterative = isinstance(c, type) and issubclass(c, UsesTrainer)
            case _:
                self.log.error("non-class component cannot be searched", component=comp_name)
                self.log.info("invalid component node: %s", comp)
                raise RuntimeError("attempted to search non-class component")
        self.pipe_name = pb.name

    @property
    def mode(self):
        if self.spec.search.metric == "RMSE":
            return "min"
        else:
            return "max"

    def set_data(
        self, train: Dataset | Path, test: ItemListCollection | Path, *, name: str | None = None
    ):
        """
        Set the data to be used for tuning.
        """
        if isinstance(train, Path):
            train = Dataset.load(train)
        if isinstance(test, Path):
            test = ItemListCollection.load_parquet(test)

        self.data = TTSplit(train, test, name=name or train.name)

    def setup(self):
        """
        Set up to run the trainer.  After this method completes, the
        :attr:`tuner` is ready.
        """
        ensure_cluster()
        self.setup_harness()
        self.tuner = self.create_tuner()
        self.out_dir.mkdir(exist_ok=True, parents=True)

    def run(self) -> ray.tune.ResultGrid:
        """
        Run the tuning job.

        Saves the results in :attr:`results`, and also returns them.
        """
        if not hasattr(self, "tuner"):
            self.setup()

        self.log.info("starting hyperparameter search")
        self.results = self.tuner.fit()
        self.log.info("finished hyperparameter search")

        return self.results

    def best_result(self, *, scope: str = "all") -> dict[str, JsonValue]:
        """
        Get the best configuration and its validation metrics.

        Args:
            scope:
                The metric search scope for iterative training.  Set to
                ``"last"`` to use the last iteration instead of the best
                iteration.  See :meth:`ray.tune.ResultGrid.get_best_result` for
                details.
        """
        best = self.results.get_best_result(scope=scope)
        res = best.metrics
        if res is None:
            raise ValueError("best result has no metrics")

        if "training_iteration" in res:
            res["config"] = res["config"] | {"epochs": res["training_iteration"]}

        return res

    def best_pipeline(self) -> PipelineConfig:
        """
        Get the (full) configuration for the best pipeline.
        """
        best = self.best_result()
        cfg = self.spec.pipeline
        assert isinstance(cfg, PipelineConfig)
        name = self.spec.component_name
        assert name is not None
        return cfg.merge_component_configs({name: best["config"]})  # type: ignore

    def search_space(self):
        """
        Get the Ray search space.
        """
        # we have exactly one
        for space in self.spec.space.values():
            return _make_space(space)

    def setup_harness(self):
        self.log.info("setting up test harness")

        self.log.info("pushing data to cluster")
        data_ref = ray.put(self.data)

        self.job = TuningJobData(
            spec=self.spec,
            random_seed=self.random_seed.spawn(1)[0],
            data_name=self.data.train.name,
            data_ref=data_ref,
        )

        if self.iterative:
            harness = ray.tune.with_parameters(IterativeEval, job=self.job)
        else:
            harness = SimplePointEval(self.job)

        paracfg = get_parallel_config()

        self.log.info(
            "setting up parallel tuner",
            cpus=paracfg.total_threads,
        )

        match self.spec.search.num_cpus:
            case "threads":
                tune_cpus = paracfg.threads
            case "all-threads":
                tune_cpus = paracfg.total_threads
            case int(n) if n > 0:
                tune_cpus = n
            case _:
                raise ValueError(f"invalid CPU count {self.spec.search.num_cpus}")

        self.harness = ray.tune.with_resources(
            harness, {"CPU": tune_cpus, "GPU": self.spec.search.num_gpus * self.settings.gpu_mult}
        )

    @property
    def metric(self):
        metric = self.spec.search.metric
        if metric is None:
            raise RuntimeError("no metric specified")
        else:
            return metric

    def create_tuner(self) -> ray.tune.Tuner:
        """
        Create a Ray tuner for the search.
        """
        match self.spec.search.method:
            case "optuna":
                return self._create_optuna_tuner()
            case "hyperopt":
                return self._create_hyperopt_tuner()
            case "random":
                return self._create_random_tuner()
            case _:
                raise ValueError(f"unsupported search method {self.spec.search.method}")

    def _create_random_tuner(self) -> ray.tune.Tuner:
        searcher = ray.tune.search.BasicVariantGenerator(
            random_state=default_rng(self.random_seed.spawn(1)[0])
        )
        return self._create_tuner_for_searcher(searcher)

    def _create_hyperopt_tuner(self) -> ray.tune.Tuner:
        searcher = HyperOptSearch(random_state_seed=int_seed(self.random_seed.spawn(1)[0]))
        return self._create_tuner_for_searcher(searcher)

    def _create_optuna_tuner(self) -> ray.tune.Tuner:
        searcher = OptunaSearch(seed=int_seed(self.random_seed.spawn(1)[0]))
        return self._create_tuner_for_searcher(searcher)

    def _create_tuner_for_searcher(self, searcher) -> ray.tune.Tuner:
        ray_store = self.out_dir / "trial-data"
        scheduler = None
        stopper = None
        cp_config = None
        if self.iterative:
            # FIXME: make this configurable
            min_iter = self.spec.search.min_epochs
            scheduler = ray.tune.schedulers.MedianStoppingRule(
                time_attr="training_iteration",
                grace_period=min_iter,
                min_time_slice=3,
                min_samples_required=3,
            )
            stopper = RelativePlateauStopper(
                metric=self.metric,
                mode=self.mode,
                grace_period=min_iter,
                check_iters=min(min_iter, 3),
                min_improvement=0.005,
            )

            cp_freq = self.spec.search.checkpoint_iters
            self.log.info("will checkpoint every %d iterations", cp_freq)
            cp_config = ray.tune.CheckpointConfig(
                checkpoint_frequency=cp_freq,
                num_to_keep=2,
                # we don't need final model checkpoints
                checkpoint_at_end=False,
            )

        nsamp = self.spec.search.max_points
        space = self.search_space()
        self.log.info("creating tuner for %d samples", nsamp, space=space)
        self.tuner = ray.tune.Tuner(
            self.harness,
            param_space=space,
            tune_config=ray.tune.TuneConfig(
                metric=self.metric,
                mode=self.mode,
                num_samples=nsamp,
                max_concurrent_trials=self.settings.jobs,
                search_alg=searcher,
                scheduler=scheduler,
            ),
            run_config=ray.tune.RunConfig(
                storage_path=ray_store.absolute().as_uri(),
                verbose=None,
                progress_reporter=ProgressReport(self.pipe_name),
                failure_config=ray.tune.FailureConfig(fail_fast=True),
                callbacks=[StatusCallback(self.pipe_name, self.data.train.name)],
                stop=stopper,
                checkpoint_config=cp_config,
            ),
        )
        return self.tuner


def _make_space(space: SearchSpace):
    out = {}
    for name, spec in space.items():
        if isinstance(spec, dict):
            out[name] = _make_space(spec)
        elif spec.type == "int" and spec.scale == "uniform":
            assert isinstance(spec.min, int)
            assert isinstance(spec.max, int)
            out[name] = ray.tune.randint(spec.min, spec.max)
        elif spec.type == "int" and spec.scale == "log":
            assert isinstance(spec.min, int)
            assert isinstance(spec.max, int)
            out[name] = ray.tune.lograndint(spec.min, spec.max, base=spec.base)
        elif spec.type == "float" and spec.scale == "uniform":
            out[name] = ray.tune.uniform(spec.min, spec.max)
        elif spec.type == "float" and spec.scale == "log":
            out[name] = ray.tune.loguniform(spec.min, spec.max, base=spec.base)

    return out



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Efficient solver routines.
"""

import numpy as np
from scipy.linalg import cho_factor, cho_solve

from lenskit.data.types import NPMatrix, NPVector


def solve_cholesky(A: NPMatrix, y: NPVector) -> NPVector:
    """
    Solve the system :math:`A\\mathbf{x}=\\mathbf{y}` for :math:`\\mathbf{x}`
    with Cholesky decomposition.

    This wraps :func:`torch.linalg.cholesky_ex` and :func:`torch.cholesky_solve`
    in an easier-to-use interface with error checking.

    Args:
        A:
            the left-hand matrix :math:`A`
        y:
            the right-hand vector :math:`\\mathbf{y}`

    Returns:
        the solution :math:`\\mathbf{x}`
    """
    if len(y.shape) > 1:  # pragma: no cover
        raise TypeError(f"y must be 1D (found shape {y.shape})")
    (n,) = y.shape
    if A.shape != (n, n):  # pragma: no cover
        raise TypeError("A must be n⨉n")

    L, low = cho_factor(A)
    return np.require(cho_solve((L, low), y), dtype=A.dtype)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import warnings

import numpy as np
from pydantic import BaseModel

from lenskit.data import ItemList
from lenskit.data.query import QueryInput, RecQuery
from lenskit.pipeline import Component
from lenskit.random import DerivableSeed, RNGFactory, derivable_rng
from lenskit.stats import argtopn


class RandomConfig(BaseModel, arbitrary_types_allowed=True):
    n: int | None = None
    """
    The number of items to select. -1 or ``None`` to return all scored items.
    """

    rng: DerivableSeed = None
    """
    Random number generator configuration.
    """


class RandomSelector(Component[ItemList]):
    """
    Randomly select items from a candidate list.

    Stability:
        Caller

    Args:
        n:
            The number of items to select, or ``-1`` to randomly permute the
            items.
        rng:
            The random number generator or specification (see :ref:`rng`).  This
            class supports derivable RNGs.
    """

    config: RandomConfig
    _rng_factory: RNGFactory

    def __init__(self, config: RandomConfig | None = None, **kwargs):
        super().__init__(config, **kwargs)
        self._rng_factory = derivable_rng(self.config.rng)

    def __call__(
        self, items: ItemList, query: QueryInput | None = None, n: int | None = None
    ) -> ItemList:
        """
        Args:
            items:
                The items from which to pick.
            query:
                The recommendation query; optional, and only consulted when the
                RNG seed is user-dependent.
            n:
                The number of items to select, overriding the configured value.
        """
        if n is None:
            n = self.config.n or -1

        query = RecQuery.create(query)
        rng = self._rng_factory(query)

        if n < 0:
            n = len(items)
        else:
            n = min(n, len(items))

        if n > 0:
            picked = rng.choice(len(items), n, replace=False)
            return items[picked]
        else:
            return items[np.zeros(0, dtype=np.int32)]


class SoftmaxRanker(Component[ItemList]):
    """
    Stochastic top-N ranking with softmax sampling.

    This uses the “softmax” sampling method, a more efficient approximation of
    Plackett-Luce sampling than even the Gumbell trick, as documented by `Tim
    Vieira`_.  It expects a scored list of input items, and samples ``n`` items,
    with selection probabilities proportional to their scores.

    .. warning::

        This ranker does not actually compute a softmax

    .. deprecated:: 2025.3

        This ranker has been replaced with the
        :class:`lenskit.stochastic.StochasticTopNRanker`. It will be removed in
        LensKit 2026.

    .. note::

        Negative scores are clamped to (approximately) zero.

    .. _`Tim Vieiera`: https://timvieira.github.io/blog/post/2019/09/16/algorithms-for-sampling-without-replacement/

    Stability:
        Caller

    Args:
        n:
            The number of items to return (-1 to return unlimited).
        rng:
            The random number generator or specification (see :ref:`rng`).  This
            class supports derivable RNGs.
    """

    config: RandomConfig
    _rng_factory: RNGFactory

    def __init__(self, config: RandomConfig | None = None, **kwargs):
        super().__init__(config, **kwargs)
        self._rng_factory = derivable_rng(self.config.rng)
        warnings.warn(
            "SoftmaxRanker is deprecated, use StochasticTopNRanker instead", DeprecationWarning, 2
        )

    def __call__(
        self, items: ItemList, query: QueryInput | None = None, n: int | None = None
    ) -> ItemList:
        query = RecQuery.create(query)
        rng = self._rng_factory(query)

        scores = items.scores()
        if scores is None:
            raise ValueError("item list must have scores")

        valid_mask = ~np.isnan(scores)
        valid_items = items[valid_mask]
        N = len(valid_items)
        if N == 0:
            return ItemList(item_ids=[], scores=[], ordered=True)

        if n is None or n < 0:
            n = self.config.n or -1

        if n < 0 or n > N:
            n = N

        keys = np.log(rng.uniform(0, 1, N))
        keys /= np.maximum(scores[valid_mask], 1.0e-10)

        picked = argtopn(keys, n)
        return ItemList(valid_items[picked], ordered=True)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import torch
from torch import Tensor, nn
from torch.linalg import norm, vecdot


class FlexMFModel(nn.Module):
    """
    Torch module that defines the flexible matrix factorization model.

    Args:
        e_size:
            The size of the embedding space.
        n_users:
            The number o fusers
        n_items:
            The number of items.
        rng:
            The Torch RNG to initialize the embeddings.
        user_bias:
            Whether to learn a user bias term.
        item_bias:
            Whether to learn an item bias term.
    """

    n_users: int
    n_items: int
    e_size: int

    u_bias: nn.Embedding | None = None
    i_bias: nn.Embedding | None = None
    u_embed: nn.Embedding
    i_embed: nn.Embedding

    def __init__(
        self,
        e_size: int,
        n_users: int,
        n_items: int,
        rng: torch.Generator,
        init_scale: float = 1.0,
        user_bias: bool = True,
        item_bias: bool = True,
        sparse: bool = False,
    ):
        super().__init__()
        self.e_size = e_size
        self.n_users = n_users
        self.n_items = n_items

        # user and item bias terms
        if user_bias:
            self.u_bias = nn.Embedding(n_users, 1, sparse=sparse)
        if item_bias:
            self.i_bias = nn.Embedding(n_items, 1, sparse=sparse)

        # user and item embeddings
        self.u_embed = nn.Embedding(n_users, e_size, sparse=sparse)
        self.i_embed = nn.Embedding(n_items, e_size, sparse=sparse)

        # initialize all values to a small normal
        if self.u_bias is not None:
            nn.init.normal_(self.u_bias.weight, std=init_scale, generator=rng)
        if self.i_bias is not None:
            nn.init.normal_(self.i_bias.weight, std=init_scale, generator=rng)

        nn.init.normal_(self.u_embed.weight, std=init_scale, generator=rng)
        nn.init.normal_(self.i_embed.weight, std=init_scale, generator=rng)

    @property
    def device(self):
        """
        Get the device.  FlexMF models do not support multiple simultaneous
        devices.
        """
        return self.i_embed.weight.device

    def zero_users(self, users: Tensor):
        """
        Zero the weights for the specified users (used to clear users who
        have no training interactions).
        """

        if self.u_bias is not None:
            self.u_bias.weight.data[users] = 0

        self.u_embed.weight.data[users, :] = 0

    def zero_items(self, items: Tensor):
        """
        Zero the weights for the specified items (used to clear items that
        have no training interactions).
        """

        if self.i_bias is not None:
            self.i_bias.weight.data[items] = 0

        self.i_embed.weight.data[items, :] = 0

    def forward(self, user: Tensor, item: Tensor, *, return_norm: bool = False):
        """
        Matrix factorization forward pass.

        This can be applied to batches of size :math:`B` or to a set of items
        with the same user. The item tensor can also be dimensonal.

        Args:
            user:
                The user number(s), typically of size :math:`1` or :math:`B`.
            items:
                The items, typically of size :math:`B` or :math:`B \\times k`
                (to score :math:`k` items for each user).
            return_norm:
                If ``True``, return the L2 norms of the parameters affecting
                each score.  In this case, the resulting tensor has an extra
                first dimension of size 2, so ``result[0]`` is the scores and
                ``result[1]`` is the norms.
        Returns:
            The scores (possibly with norms).
        """

        # look up biases and embeddings
        zero = torch.tensor(0.0)
        ub = self.u_bias(user).reshape(user.shape) if self.u_bias is not None else zero
        ib = self.i_bias(item).reshape(item.shape) if self.i_bias is not None else zero

        uvec = self.u_embed(user)
        ivec = self.i_embed(item)

        # compute the inner score
        ips = vecdot(uvec, ivec)
        score = ub + ib + ips

        if return_norm:
            l2 = torch.square(ub) + torch.square(ib) + norm(uvec, dim=-1) + norm(ivec, dim=-1)
            assert l2.shape == score.shape
            return torch.stack((score, l2))
        else:
            # we're done
            return score



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path

from lenskit.config import LenskitSettings, configure, lenskit_config


def test_default_config():
    cfg = lenskit_config()
    assert cfg is not None
    assert cfg.random.seed is None


def test_load_toml():
    cfg = configure(cfg_dir=Path(__file__).parent, _set_global=False)  # type: ignore
    assert cfg is not None
    assert cfg.random.seed == 42
    assert set(cfg.machines.keys()) == {"local", "shared"}



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import pickle

from lenskit.data import Dataset


def test_dataset_pickle(ml_ds: Dataset):
    data = pickle.dumps(ml_ds)

    ds = pickle.loads(data)
    assert isinstance(ds, Dataset)

    assert ds.item_count == ml_ds.item_count
    assert ds.user_count == ml_ds.user_count
    assert ds.interaction_count == ml_ds.interaction_count



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import json
from os import fspath
from pathlib import Path

from click.testing import CliRunner

from lenskit.cli import lenskit
from lenskit.testing import ml_test_dir


def test_data_convert(tmpdir: Path):
    out_path = tmpdir / "ml-data"
    schema_file = out_path / "schema.json"

    runner = CliRunner()
    result = runner.invoke(
        lenskit, ["data", "convert", "--movielens", fspath(ml_test_dir), fspath(out_path)]
    )

    assert result.exit_code == 0
    assert schema_file.exists()

    data = json.loads(schema_file.read_text("utf8"))
    assert data["name"] == "ml-latest-small"
